# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for BIND server."""

import pathlib
import re
from collections import defaultdict
from pathlib import Path

import augeas

from plinth import action_utils
from plinth.actions import privileged

CONFIG_FILE = '/etc/bind/named.conf.options'
ZONES_DIR = '/var/bind/pri'

DEFAULT_CONFIG = '''
acl goodclients {
    localnets;
};
options {
listen-on { !10.42.0.1; !10.42.1.1; !10.42.2.1; !10.42.3.1; !10.42.4.1; !10.42.5.1; !10.42.6.1; !10.42.7.1; any; };
directory "/var/cache/bind";

recursion yes;
allow-query { goodclients; };

forwarders {
127.0.0.53;
};
forward first;

auth-nxdomain no;    # conform to RFC1035
listen-on-v6 { any; };
};
'''  # noqa: E501
DEFAULT_FORWARDER = '127.0.0.53'  # systemd-resolved
LISTEN_ON = 'listen-on { !10.42.0.1; !10.42.1.1; !10.42.2.1; !10.42.3.1; '\
    '!10.42.4.1; !10.42.5.1; !10.42.6.1; !10.42.7.1; any; };'


@privileged
def setup(old_version: int):
    """Setup BIND configuration."""
    if old_version == 0:
        with open(CONFIG_FILE, 'w', encoding='utf-8') as conf_file:
            conf_file.write(DEFAULT_CONFIG)
    elif old_version < 4:
        if not get_config()['forwarders']:
            _set_forwarders(DEFAULT_FORWARDER)

        if not _has_listen_on():
            _insert_listen_on()

        if old_version < 3:
            _remove_dnssec()

    Path(ZONES_DIR).mkdir(exist_ok=True, parents=True)

    action_utils.service_try_restart('named')


@privileged
def configure(forwarders: str):
    """Configure BIND."""
    _set_forwarders(forwarders)
    action_utils.service_try_restart('named')


def get_config():
    """Get current configuration."""
    data = [line.strip() for line in open(CONFIG_FILE, 'r', encoding='utf-8')]

    forwarders = ''
    flag = False
    for line in data:
        if re.match(r'^\s*forwarders\s+{', line):
            flag = True
        elif flag and '//' not in line:
            forwarders = re.sub('[;]', '', line)
            flag = False

    return {'forwarders': forwarders}


def _set_forwarders(forwarders):
    """Set DNS forwarders."""
    flag = 0
    data = [line.strip() for line in open(CONFIG_FILE, 'r', encoding='utf-8')]
    conf_file = open(CONFIG_FILE, 'w', encoding='utf-8')
    for line in data:
        if re.match(r'^\s*forwarders\s+{', line):
            conf_file.write(line + '\n')
            for dns in forwarders.split():
                conf_file.write(dns + '; ')
            conf_file.write('\n')
            flag = 1
        elif '};' in line and flag == 1:
            conf_file.write(line + '\n')
            flag = 0
        elif flag == 0:
            conf_file.write(line + '\n')
    conf_file.close()


def _remove_dnssec():
    """Remove DNSSEC options."""
    data = [line.strip() for line in open(CONFIG_FILE, 'r', encoding='utf-8')]

    with open(CONFIG_FILE, 'w', encoding='utf-8') as file_handle:
        for line in data:
            if not re.match(r'^\s*dnssec-enable\s+yes;', line):
                file_handle.write(line + '\n')


def _has_listen_on():
    """Return whether listen-on config option is present."""
    lines = pathlib.Path(CONFIG_FILE).read_text().splitlines()
    regex = r'^\s*listen-on\s+{'
    return any((re.match(regex, line) for line in lines))


def _insert_listen_on():
    """Insert the listen-on option."""
    config_file = pathlib.Path(CONFIG_FILE)
    lines = config_file.read_text().splitlines(keepends=True)
    write_lines = []
    for line in lines:
        write_lines += line
        if re.match(r'^\s*options\s+{', line):
            write_lines += LISTEN_ON + '\n'

    config_file.write_text(''.join(write_lines))


def get_served_domains():
    """Return list of domains service handles.

    Augeas path for zone files:
    ===========================
    augtool> print /files/var/bind/pri/local.zone
    /files/var/bind/pri/local.zone
    /files/var/bind/pri/local.zone/$TTL = "604800"
    /files/var/bind/pri/local.zone/@[1]
    /files/var/bind/pri/local.zone/@[1]/1
    /files/var/bind/pri/local.zone/@[1]/1/class = "IN"
    /files/var/bind/pri/local.zone/@[1]/1/type = "SOA"
    /files/var/bind/pri/local.zone/@[1]/1/mname = "localhost."
    /files/var/bind/pri/local.zone/@[1]/1/rname = "root.localhost."
    /files/var/bind/pri/local.zone/@[1]/1/serial = "2"
    /files/var/bind/pri/local.zone/@[1]/1/refresh = "604800"
    /files/var/bind/pri/local.zone/@[1]/1/retry = "86400"
    /files/var/bind/pri/local.zone/@[1]/1/expiry = "2419200"
    /files/var/bind/pri/local.zone/@[1]/1/minimum = "604800"
    /files/var/bind/pri/local.zone/@[2]
    /files/var/bind/pri/local.zone/@[2]/1
    /files/var/bind/pri/local.zone/@[2]/1/class = "IN"
    /files/var/bind/pri/local.zone/@[2]/1/type = "NS"
    /files/var/bind/pri/local.zone/@[2]/1/rdata = "localhost."
    /files/var/bind/pri/local.zone/@[3]
    /files/var/bind/pri/local.zone/@[3]/1
    /files/var/bind/pri/local.zone/@[3]/1/class = "IN"
    /files/var/bind/pri/local.zone/@[3]/1/type = "A"
    /files/var/bind/pri/local.zone/@[3]/1/rdata = "127.0.0.1"
    /files/var/bind/pri/local.zone/@[4]
    /files/var/bind/pri/local.zone/@[4]/1
    /files/var/bind/pri/local.zone/@[4]/1/class = "IN"
    /files/var/bind/pri/local.zone/@[4]/1/type = "AAAA"
    /files/var/bind/pri/local.zone/@[4]/1/rdata = "::1"

    Need to find the related functionality to parse the A records

    Retrieve from /etc/bind/db* zone files all the configured A records.
    Assuming zones files in ZONES_DIR are all used.
    :return: dictionary in the form 'domain_name': ['ip_address', 'ipv6_addr']
    """
    RECORD_TYPES = ('A', 'AAAA')
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Dns_Zone/lens', 'Dns_Zone.lns')

    zone_file_path = Path(ZONES_DIR)
    zone_files = [zf for zf in zone_file_path.iterdir() if zf.is_file()]

    # augeas load only required files
    for zone_file in zone_files:
        aug.set('/augeas/load/Dns_Zone/incl[last() + 1]', str(zone_file))

    aug.load()

    served_domains = defaultdict(list)
    for zone_file in zone_files:
        base_path = '/files/%s/@[{record_order}]/1/{field}' % zone_file
        count = 1
        mname = aug.get(base_path.format(record_order=count, field='mname'))
        while True:
            record_type = aug.get(
                base_path.format(record_order=count, field='type'))

            # no record type ends the search
            if record_type is None:
                break

            if record_type in RECORD_TYPES:
                served_domains[mname].append(
                    aug.get(base_path.format(record_order=count,
                                             field='rdata')))

            count += 1

    return served_domains



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for Miniflux feed reader."""

import json
import os
import pathlib
from typing import Tuple

import pexpect

from plinth import action_utils
from plinth.actions import privileged, secret_str
from plinth.db import postgres, dbconfig
from plinth.utils import is_non_empty_file

STATIC_SETTINGS = {
    'BASE_URL': 'http://localhost/miniflux/',
    'RUN_MIGRATIONS': 1,
    'PORT': 8788
}

ENV_VARS_FILE = '/etc/miniflux/freedombox.conf'
DBCONFIG_PATH = '/etc/dbconfig-common/miniflux.conf'
DB_BACKUP_FILE = '/var/lib/plinth/backups-data/miniflux-database.sql'


def _dict_to_env_file(dictionary: dict[str, str]) -> str:
    """Write a dictionary into a systemd environment file format."""
    return "\n".join((f"{k}={v}" for k, v in dictionary.items()))


def _env_file_to_dict(env_vars: str) -> dict[str, str]:
    """Return systemd environtment variables as a dictionary."""
    return {
        line.split('=')[0]: line.split('=')[1].strip()
        for line in env_vars.splitlines()
        if line.strip() and not line.strip().startswith('#')
    }


@privileged
def pre_setup():
    """Perform pre-install actions for Miniflux."""
    vars_file = pathlib.Path(ENV_VARS_FILE)
    vars_file.parent.mkdir(parents=True, exist_ok=True)

    existing_settings = {}
    if is_non_empty_file(ENV_VARS_FILE):
        # Any comments in the file will be dropped.
        existing_settings = _env_file_to_dict(vars_file.read_text())

    new_settings = existing_settings | STATIC_SETTINGS
    vars_file.write_text(_dict_to_env_file(new_settings))


def _run_miniflux_interactively(command: str, username: str,
                                password: str) -> Tuple[str, dict]:
    """Fill interactive terminal prompt for username and password."""
    args = ['-c', '/etc/miniflux/miniflux.conf', command]
    os.environ['LOG_FORMAT'] = 'json'
    child = pexpect.spawn('miniflux', args, env=os.environ)

    # The CLI is in English only.
    child.expect('Enter Username: ')
    child.sendline(username)

    child.expect('Enter Password: ')
    child.sendline(password)

    child.expect(pexpect.EOF)
    raw_message = child.before.decode()  # type: ignore
    try:
        json_message = json.loads(raw_message)
    except (KeyError, json.JSONDecodeError):
        json_message = {}

    child.close()
    if child.exitstatus or child.signalstatus:
        message = json_message.get('msg') if json_message else raw_message
        raise Exception(message)

    return raw_message, json_message


@privileged
def create_admin_user(username: str, password: secret_str):
    """Create a new admin user for Miniflux CLI.

    Raise exception if a user with the name already exists or otherwise fails.
    """
    _, json_message = _run_miniflux_interactively('--create-admin', username,
                                                  password)
    # user_id is allocated only when a new user is created successfully.
    if json_message and not json_message.get('user_id'):
        raise Exception(json_message.get('msg'))


@privileged
def reset_user_password(username: str, password: secret_str):
    """Reset a user password using Miniflux CLI.

    Raise exception if the user does not exist or otherwise fails.
    """
    _run_miniflux_interactively('--reset-password', username, password)


@privileged
def uninstall():
    """Ensure that the database is removed."""
    action_utils.debconf_set_selections([
        'miniflux miniflux/purge boolean true',
        'miniflux miniflux/dbconfig-install boolean true',
        'miniflux miniflux/dbconfig-reinstall boolean true'
        'miniflux miniflux/dbconfig-upgrade boolean true',
        'miniflux miniflux/dbconfig-remove boolean true',
    ])


@privileged
def dump_database():
    """Dump database to file."""
    config = dbconfig.get_credentials(DBCONFIG_PATH)
    postgres.dump_database(DB_BACKUP_FILE, config['database'])


@privileged
def restore_database():
    """Restore database from file."""
    config = dbconfig.get_credentials(DBCONFIG_PATH)
    postgres.restore_database(DB_BACKUP_FILE, config['database'],
                              config['user'], config['password'])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for infinoted.
"""

from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('infinoted is a server for Gobby, a collaborative text editor.'),
    format_lazy(
        _('To use it, <a href="https://gobby.github.io/">download Gobby</a>, '
          'desktop client and install it. Then start Gobby and select '
          '"Connect to Server" and enter your {box_name}\'s domain name.'),
        box_name=_(cfg.box_name)),
]


class InfinotedApp(app_module.App):
    """FreedomBox app for infinoted."""

    app_id = 'infinoted'

    _version = 3

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('infinoted'), icon_filename='infinoted',
                               description=_description,
                               manual_page='Infinoted',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-infinoted', info.name, info.icon_filename,
                              info.tags, 'infinoted:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut(
            'shortcut-infinoted', info.name, icon=info.icon_filename,
            description=info.description, manual_page=info.manual_page,
            configure_url=reverse_lazy('infinoted:index'),
            clients=info.clients, tags=info.tags, login_required=False)
        self.add(shortcut)

        packages = Packages('packages-infinoted', ['infinoted'])
        self.add(packages)

        firewall = Firewall('firewall-infinoted', info.name,
                            ports=['infinoted-plinth'], is_external=True)
        self.add(firewall)

        daemon = Daemon('daemon-infinoted', 'infinoted',
                        listen_ports=[(6523, 'tcp4'), (6523, 'tcp6')])
        self.add(daemon)

        backup_restore = BackupRestore('backup-restore-infinoted',
                                       **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        privileged.setup()
        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        super().uninstall()
        privileged.uninstall()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for WireGuard."""

from plinth import action_utils
from plinth.actions import privileged

SERVER_INTERFACE = 'wg0'


@privileged
def get_info() -> dict[str, dict]:
    """Return info for each configured interface."""
    output = action_utils.run(['wg', 'show', 'all', 'dump'],
                              check=True).stdout.decode().strip()
    lines = output.split('\n')
    interfaces: dict[str, dict] = {}
    for line in lines:
        if not line:
            continue

        fields: list = [
            field if field != '(none)' else None for field in line.split()
        ]
        interface_name = fields[0]
        if interface_name in interfaces:
            latest_handshake = int(fields[5]) if int(fields[5]) else None
            peer = {
                'public_key': fields[1],
                'preshared_key': fields[2],
                'endpoint': fields[3],
                'allowed_ips': fields[4],
                'latest_handshake': latest_handshake,
                'transfer_rx': fields[6],
                'transfer_tx': fields[7],
                'persistent_keepalive': fields[8],
            }
            interfaces[interface_name]['peers'].append(peer)

        else:
            interfaces[interface_name] = {
                'interface_name': interface_name,
                'private_key': fields[1],
                'public_key': fields[2],
                'listen_port': fields[3],
                'fwmark': fields[4],
                'peers': [],
            }

    return interfaces



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Django signals emitted within FreedomBox.
"""

from django.dispatch import Signal

# Arguments: -
pre_module_loading = Signal()

# Arguments: -
post_app_loading = Signal()

# Arguments: module_name
post_setup = Signal()

# Arguments: old_hostname, new_hostname
pre_hostname_change = Signal()

# Arguments: old_hostname, new_hostname
post_hostname_change = Signal()

# Arguments: domain_type, name, description, services
domain_added = Signal()

# Arguments: domain_type, name
domain_removed = Signal()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Home Assistant."""

from django.utils.functional import lazy
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.container import Container
from plinth.modules.apache.components import WebserverRoot
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.package import Packages
from plinth.utils import format_lazy

from . import manifest, privileged

_alert = '''
<div class="alert alert-warning d-flex align-items-center" role="alert">
  <div class="me-2">
    <span class="fa fa-exclamation-triangle" aria-hidden="true"></span>
    <span class="visually-hidden">{}</span>
  </div>
  <div>{}</div>
</div>
'''

_description = [
    _('Home Assistant is a home automation hub with emphasis on local control '
      'and privacy. It integrates with thousands of devices including smart '
      'bulbs, alarms, presence sensors, door bells, thermostats, irrigation '
      'timers, energy monitors, etc.'),
    _('Home Assistant can detect, configure, and use various devices on the '
      'local network. For devices using other protocols such as ZigBee, it '
      'typically requires additional hardware such as a ZigBee USB dongle. '
      'You need to re-run setup if such hardware is added or removed.'),
    _('Home Assistant web interface must be setup soon after the app is '
      'installed. An administrator account is created at this time. Home '
      'Assistant maintains its own user accounts.'),
    format_lazy(
        _('Please note that Home Assistant is installed and run inside a '
          'container provided by the Home Assistant project. Security, '
          'quality, privacy and legal reviews are done by the upstream '
          'project and not by Debian/{box_name}. Updates are performed '
          'following an independent cycle.'), box_name=_(cfg.box_name)),
    format_lazy(_alert, _('Caution:'), _('This app is experimental.')),
]


class HomeAssistnatApp(app_module.App):
    """FreedomBox app for Home Assistant."""

    app_id = 'homeassistant'

    _version = 1

    def __init__(self):
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Home Assistant'),
                               icon_filename='homeassistant',
                               description=_description,
                               manual_page='HomeAssistant',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-homeassistant', info.name,
                              info.icon_filename, info.tags,
                              'homeassistant:index', parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-homeassistant', info.name,
                                      icon=info.icon_filename, url='#',
                                      clients=info.clients, tags=info.tags,
                                      login_required=True)
        self.add(shortcut)

        packages = Packages('packages-homeassistant', ['podman'],
                            conflicts=['libpam-tmpdir'],
                            conflicts_action=Packages.ConflictsAction.REMOVE)
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-homeassistant', [
            '/etc/apache2/includes/home-assistant-freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-homeassistant', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-homeassistant', ['8123'])
        self.add(firewall_local_protection)

        webserver = WebserverRoot('webserverroot-homeassistant',
                                  'home-assistant-freedombox')
        self.add(webserver)

        image_name = 'registry.freedombox.org/' \
            'homeassistant/home-assistant:stable'
        volume_name = 'home-assistant-freedombox'
        volumes = {
            '/run/dbus': '/run/dbus',
            volume_name: '/config',
        }
        devices = {
            f'/dev/ttyUSB{number}': f'/dev/ttyUSB{number}'
            for number in range(8)
        }
        container = Container(
            'container-homeassistant', 'home-assistant-freedombox',
            image_name=image_name, volume_name=volume_name,
            volume_path='/var/lib/home-assistant-freedombox/config/',
            volumes=volumes, devices=devices, listen_ports=[(8123, 'tcp4')])
        self.add(container)

        backup_restore = BackupRestore('backup-restore-homeassistant',
                                       **manifest.backup)
        self.add(backup_restore)

    def post_init(self):
        """Perform post initialization operations."""
        root = self.get_component('webserverroot-homeassistant')

        def get_url():
            return f'https://{root.domain_get()}'

        url = lazy(get_url, str)()
        self.get_component('shortcut-homeassistant').url = url
        self.info.clients[0]['platforms'][0]['url'] = url

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)

        privileged.setup()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure OpenVPN server."""

import datetime
import os
import pathlib
import shutil
import subprocess

import augeas

from plinth import action_utils
from plinth.actions import privileged

KEYS_DIRECTORY = pathlib.Path('/etc/openvpn/freedombox-keys')
CA_CERTIFICATE_PATH = KEYS_DIRECTORY / 'pki' / 'ca.crt'
USER_CERTIFICATE_PATH = KEYS_DIRECTORY / 'pki' / 'issued' / '{username}.crt'
USER_KEY_PATH = KEYS_DIRECTORY / 'pki' / 'private' / '{username}.key'
ATTR_FILE = KEYS_DIRECTORY / 'pki' / 'index.txt.attr'

SERVER_CONFIGURATION_PATH = '/etc/openvpn/server/freedombox.conf'

SERVICE_NAME = 'openvpn-server@freedombox'

SERVER_CONFIGURATION = '''
port 1194
proto udp
proto udp6
dev tun

client-to-client

ca /etc/openvpn/freedombox-keys/pki/ca.crt
cert /etc/openvpn/freedombox-keys/pki/issued/server.crt
key /etc/openvpn/freedombox-keys/pki/private/server.key

dh none

server 10.91.0.0 255.255.255.0
keepalive 10 120
verb 3

tls-server
tls-version-min 1.2
cipher AES-256-CBC
script-security 2
'''

CLIENT_CONFIGURATION = '''
client
remote {remote} 1194
proto udp
proto udp6
dev tun
nobind
remote-cert-tls server
cipher AES-256-CBC
redirect-gateway
verb 3
<ca>
{ca}</ca>
<cert>
{cert}</cert>
<key>
{key}</key>'''

_EASY_RSA_CONFIGURATION = {
    'EASYRSA_ALGO': 'ec',  # Use Elliptic Curve Cryptography by default
    'EASYRSA_BATCH': '1',  # Prevent prompting
    'EASYRSA_DIGEST': 'sha512',
    'EASYRSA_CA_EXPIRE': '3650',  # 10 years expiry for CA root certificate
    'EASYRSA_CERT_EXPIRE': '3650',  # 10 years expiry for server/client certs
    'EASYRSA_CERT_RENEW': '1095',  # Renew cert if expiry less than 3 years
    'EASYRSA_REQ_COUNTRY': 'US',
    'EASYRSA_REQ_PROVINCE': 'NY',
    'EASYRSA_REQ_CITY': 'New York',
    'EASYRSA_REQ_ORG': 'FreedomBox',
    'EASYRSA_REQ_EMAIL': 'me@freedombox',
    'EASYRSA_REQ_OU': 'Home',
}


@privileged
def setup():
    """Setup configuration, CA and certificates."""
    _write_server_config()
    _create_certificates()
    _setup_firewall()
    action_utils.service_try_restart(SERVICE_NAME)


def _write_server_config():
    """Write server configuration."""
    with open(SERVER_CONFIGURATION_PATH, 'w', encoding='utf-8') as file_handle:
        file_handle.write(SERVER_CONFIGURATION)


def _setup_firewall():
    """Add TUN device to internal zone in firewalld."""

    def _configure_interface(interface, operation):
        """Add or remove an interface into internal zone."""
        command = [
            'firewall-cmd', '--zone', 'internal',
            '--{}-interface'.format(operation), interface
        ]
        action_utils.run(command, check=False)
        action_utils.run(command + ['--permanent'], check=False)

    def _is_tunplus_enabled():
        """Return whether tun+ interface is already added."""
        try:
            process = action_utils.run(
                ['firewall-cmd', '--zone', 'internal', '--list-interfaces'],
                check=True)
            return 'tun+' in process.stdout.decode().strip().split()
        except subprocess.CalledProcessError:
            return True  # Safer

    # XXX: Due to https://bugs.debian.org/919517 when tun+ interface is added,
    # firewalld is unable to handle it in nftables backend causing firewalld to
    # break while applying rules. This makes the entire system unreachable.
    # Hack around the problem by adding a few tun interfaces into the internal
    # zone. Hopefully, OpenVPN setting 'dev tun' will end up using one of those
    # if the tun devices are not used by other services. When the issue is
    # fixed, use tun+ instead.
    is_tunplus_set = _is_tunplus_enabled()
    _configure_interface('tun+', 'remove')
    for index in range(8):
        _configure_interface('tun{}'.format(index), 'add')

    if is_tunplus_set:
        action_utils.service_restart('firewalld')


def _run_easy_rsa(args):
    """Execute easy-rsa command with some default arguments."""
    return action_utils.run(['/usr/share/easy-rsa/easyrsa'] + args,
                            cwd=KEYS_DIRECTORY, check=True)


def _write_easy_rsa_config():
    """Write easy-rsa 'vars' file."""
    with (KEYS_DIRECTORY / 'pki' / 'vars').open('w') as file_handle:
        for key, value in _EASY_RSA_CONFIGURATION.items():
            file_handle.write(f'set_var {key} "{value}"\n')


def _is_renewable(cert_name):
    """Return whether a certificate is within configured renewable days.

    'easy-rsa renewable' command could be used to perform the check. However,
    the script fetches the expiry date of the certificate from the index.txt
    file. When this file has multiple entries for the same certificate base
    name, the results of the command are undesirable. Multiple entries for the
    same certificate base name can occur in the index.txt file in some unusual
    cases. For example, earlier versions of FreedomBox ran build-server-full
    followed by gen-req/sign-req. This approach created such entries. So,
    determine expiry here without using easy-rsa script.
    """
    cert_path = KEYS_DIRECTORY / 'pki' / 'issued' / (cert_name + '.crt')
    if not cert_path.exists():
        return False

    process = action_utils.run(
        ['openssl', 'x509', '-noout', '-enddate', '-in',
         str(cert_path)], check=True)
    date_string = process.stdout.decode().strip().partition('=')[2]
    cert_expiry_time = datetime.datetime.strptime(date_string,
                                                  '%b %d %H:%M:%S %Y GMT')
    cert_expiry_time = cert_expiry_time.replace(tzinfo=datetime.timezone.utc)

    now = datetime.datetime.now(tz=datetime.timezone.utc)
    renew_period = datetime.timedelta(
        days=int(_EASY_RSA_CONFIGURATION['EASYRSA_CERT_RENEW']))
    return (cert_expiry_time - now) < renew_period


def _renew(cert_name):
    """Renew a certificate and revoke the old certificate.

    Without revoking the old certificate, another renewal is not possible due
    to safety checks in easy-rsa script.
    """
    _run_easy_rsa(['renew', cert_name, 'nopass'])

    # Remove the old certificate so that more renewals can work
    _run_easy_rsa(['revoke-renewed', cert_name])


def _create_certificates():
    """Generate CA and server certificates."""
    KEYS_DIRECTORY.mkdir(mode=0o700, exist_ok=True)

    # Don't re-initialize PKI if it already exists. This will lead to wiping of
    # all existing certificates and downloaded profiles.
    if not (KEYS_DIRECTORY / 'pki').is_dir():
        _run_easy_rsa(['init-pki'])

    _write_easy_rsa_config()

    # Don't reinitialize the CA certificates. This will invalidate all existing
    # server/client certificates and downloaded client profiles.
    if not CA_CERTIFICATE_PATH.exists():
        _run_easy_rsa(['build-ca', 'nopass'])

    # Renew server certificate if already exists. Already downloaded profiles
    # don't change.
    server_cert = KEYS_DIRECTORY / 'pki' / 'issued' / 'server.crt'
    if not server_cert.exists():
        _run_easy_rsa(['build-server-full', 'server', 'nopass'])
    elif _is_renewable('server'):
        _renew('server')


@privileged
def get_profile(username: str, remote_server: str) -> str:
    """Return the profile for a user."""
    if username == 'ca' or username == 'server':
        raise Exception('Invalid username')

    user_certificate = str(USER_CERTIFICATE_PATH).format(username=username)
    user_key = str(USER_KEY_PATH).format(username=username)

    if not _is_non_empty_file(user_certificate) or \
       not _is_non_empty_file(user_key):
        set_unique_subject('no')  # Set unique subject in attribute file to no
        _run_easy_rsa(['build-client-full', username, 'nopass'])
    elif _is_renewable(username):
        _renew(username)

    user_certificate_string = _read_file(user_certificate)
    user_key_string = _read_file(user_key)
    ca_string = _read_file(CA_CERTIFICATE_PATH)

    return CLIENT_CONFIGURATION.format(ca=ca_string,
                                       cert=user_certificate_string,
                                       key=user_key_string,
                                       remote=remote_server)


def set_unique_subject(value):
    """Set the unique_subject value to a particular value."""
    aug = load_augeas()
    aug.set('/files' + str(ATTR_FILE) + '/unique_subject', value)
    aug.save()


def _read_file(filename):
    """Return the entire contents of a file as string."""
    with open(filename, 'r', encoding='utf-8') as file_handle:
        return ''.join(file_handle.readlines())


def _is_non_empty_file(filepath):
    """Return whether a file exists and is not zero size."""
    return os.path.isfile(filepath) and os.path.getsize(filepath) > 0


def load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)

    # shell-script config file lens
    aug.set('/augeas/load/Simplevars/lens', 'Simplevars.lns')
    aug.set('/augeas/load/Simplevars/incl[last() + 1]', str(ATTR_FILE))
    aug.load()
    return aug


@privileged
def uninstall():
    """Remove configuration directory for OpenVPN."""
    shutil.rmtree('/etc/openvpn', ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure searx."""

import gzip
import os
import pathlib
import secrets
import shutil

import augeas
import yaml

from plinth import action_utils
from plinth.actions import privileged
from plinth.modules.searx.manifest import PUBLIC_ACCESS_SETTING_FILE
from plinth.utils import gunzip

SETTINGS_FILE = '/etc/searx/settings.yml'

UWSGI_FILE = '/etc/uwsgi/apps-available/searx.ini'


def _copy_uwsgi_configuration():
    """Copy example uwsgi configuration

    Copy the example uwsgi configuration shipped with Searx documentation to
    the appropriate uwsgi directory.
    """
    example_config = ('/usr/share/doc/searx/examples/'
                      'uwsgi/apps-available/searx.ini')
    if not os.path.exists(UWSGI_FILE):
        shutil.copy(example_config, os.path.dirname(UWSGI_FILE))


def _update_uwsgi_configuration():
    """Fix uwsgi configuration.

    uwsgi 2.0.15-debian crashes when trying to autoload.
    """
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/inifile/lens', 'Puppet.lns')
    aug.set('/augeas/load/inifile/incl[last() + 1]', UWSGI_FILE)
    aug.load()
    aug.set('/files/etc/uwsgi/apps-available/searx.ini/uwsgi/autoload',
            'false')
    aug.set('/files/etc/uwsgi/apps-available/searx.ini/uwsgi/chmod-socket',
            '660')
    aug.save()


def _generate_secret_key(settings):
    """Generate a secret key for the Searx installation."""
    secret_key = secrets.token_hex(64)
    settings['server']['secret_key'] = secret_key


def _set_title(settings):
    """Set the page title to '{box_name} Web Search'."""
    title = 'FreedomBox Web Search'
    settings['general']['instance_name'] = title


def _set_timeout(settings):
    """Set timeout to 20 seconds."""
    settings['outgoing']['request_timeout'] = 20.0


def _set_safe_search(settings):
    """Set safe search to Moderate."""
    settings['search']['safe_search'] = 1


@privileged
def set_safe_search(filter_: int):
    """Set safe search filter for search results."""
    settings = read_settings()
    settings['search']['safe_search'] = filter_
    write_settings(settings)


@privileged
def get_safe_search() -> int:
    """Return the value of the safe search setting."""
    if os.path.exists(SETTINGS_FILE):
        settings = read_settings()
        return int(settings['search']['safe_search'])
    else:
        return 0


def read_settings():
    """Load settings as dictionary from YAML config file."""
    with open(SETTINGS_FILE, 'rb') as settings_file:
        return yaml.safe_load(settings_file)


def write_settings(settings):
    """Write settings from dictionary to YAML config file."""
    with open(SETTINGS_FILE, 'w', encoding='utf-8') as settings_file:
        yaml.dump(settings, settings_file)


def _get_example_settings_file():
    searx_doc_dir = pathlib.Path('/usr/share/doc/searx/examples/')
    if (searx_doc_dir / 'settings.yml').exists():
        return searx_doc_dir / 'settings.yml'

    return searx_doc_dir / 'settings.yml.gz'


def _update_search_engines(settings):
    """Update settings with the latest supported search engines."""
    example_settings_file = _get_example_settings_file()
    open_func = gzip.open if example_settings_file.suffix == '.gz' else open
    with open_func(example_settings_file, 'rb') as example_settings:
        settings['engines'] = yaml.safe_load(example_settings)['engines']


@privileged
def setup():
    """Post installation actions for Searx."""
    _copy_uwsgi_configuration()
    _update_uwsgi_configuration()

    if not os.path.exists(SETTINGS_FILE):
        example_settings_file = _get_example_settings_file()
        if example_settings_file.suffix == '.gz':
            gunzip(str(example_settings_file), SETTINGS_FILE)
        else:
            pathlib.Path(SETTINGS_FILE).parent.mkdir(mode=0o755)
            shutil.copy(example_settings_file, SETTINGS_FILE)

    settings = read_settings()
    _generate_secret_key(settings)
    _set_title(settings)
    _set_timeout(settings)
    _set_safe_search(settings)
    _update_search_engines(settings)
    write_settings(settings)

    # Service is started again by socket.
    action_utils.service_stop('uwsgi-app@searx.service')


@privileged
def enable_public_access():
    """Enable public access to the SearX application."""
    open(PUBLIC_ACCESS_SETTING_FILE, 'w', encoding='utf-8').close()


@privileged
def disable_public_access():
    """Disable public access to the SearX application."""
    if os.path.exists(PUBLIC_ACCESS_SETTING_FILE):
        os.remove(PUBLIC_ACCESS_SETTING_FILE)


@privileged
def uninstall():
    """Remove configuration uWSGI file."""
    shutil.rmtree('/etc/searx', ignore_errors=True)
    pathlib.Path(UWSGI_FILE).unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for Miniflux."""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon, SharedDaemon
from plinth.modules.apache.components import Webserver
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import Firewall
from plinth.package import Packages

from . import manifest, privileged

_description = [
    _('Miniflux is a web-based tool that aggregates news and blog updates from'
      ' various websites into one centralized, easy-to-read format. It has a '
      'simple interface and focuses on a distraction-free reading experience. '
      'You can can subscribe to your favorite sites and access full article '
      'contents within the reader itself.'),
    _('Key features include keyboard shortcuts for quick navigation, full-text'
      ' search, filtering articles, categories and favorites. Miniflux '
      'preserves user privacy by removing trackers. The primary interface is '
      'web-based. There are several third-party '
      '<a href="https://miniflux.app/docs/apps.html">clients</a> as well.'),
]


class MinifluxApp(app_module.App):
    """FreedomBox app for Miniflux."""

    app_id = 'miniflux'

    _version = 2

    def __init__(self):
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(self.app_id, self._version, name=_('Miniflux'),
                               icon_filename='miniflux',
                               description=_description,
                               manual_page='miniflux',
                               clients=manifest.clients, tags=manifest.tags,
                               donation_url='https://miniflux.app/#donations')
        self.add(info)

        menu_item = menu.Menu('menu-miniflux', info.name, info.icon_filename,
                              info.tags, 'miniflux:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-miniflux', info.name,
                                      info.icon_filename, url='/miniflux',
                                      clients=manifest.clients, tags=info.tags,
                                      login_required=True)
        self.add(shortcut)

        # miniflux package does not depend on postgres. Install postgresql,
        # start it and only then install miniflux.
        packages = Packages('packages-miniflux-postgresql', ['postgresql'])
        self.add(packages)

        packages = Packages('packages-miniflux', ['miniflux'])
        self.add(packages)

        drop_in_configs = DropinConfigs(
            'dropin-configs-miniflux',
            ['/etc/apache2/conf-available/miniflux-freedombox.conf'])
        self.add(drop_in_configs)

        firewall = Firewall('firewall-miniflux', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        webserver = Webserver('webserver-miniflux', 'miniflux-freedombox',
                              urls=['https://{host}/miniflux/'])
        self.add(webserver)

        daemon = SharedDaemon('shared-daemon-miniflux-postgresql',
                              'postgresql')
        self.add(daemon)

        daemon = Daemon('daemon-miniflux', 'miniflux',
                        listen_ports=[(8788, 'tcp4'), (8788, 'tcp6')])
        self.add(daemon)

        backup_restore = MinifluxBackupRestore('backup-restore-miniflux',
                                               **manifest.backup)
        self.add(backup_restore)

    def setup(self, old_version=None):
        """Install and configure the app."""
        privileged.pre_setup()

        # Database needs to be running for successful initialization or upgrade
        # of miniflux database. 1. The app and database are being freshly
        # installed. 2. The app is being freshly installed, but database server
        # is already installed but disabled. 3. The app is being updated to new
        # version but is currently disabled (DB is installed but disabled or
        # enabled).
        with self.get_component(
                'shared-daemon-miniflux-postgresql').ensure_running():
            super().setup(old_version)

        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        privileged.uninstall()
        with self.get_component(
                'shared-daemon-miniflux-postgresql').ensure_running():
            # Database needs to be running for successful removal miniflux
            # database.
            super().uninstall()


class MinifluxBackupRestore(BackupRestore):
    """Component to backup/restore Miniflux."""

    def backup_pre(self, packet):
        """Save database contents."""
        super().backup_pre(packet)
        privileged.dump_database()

    def restore_post(self, packet):
        """Restore database contents."""
        super().restore_post(packet)
        privileged.restore_database()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Apache web server."""

import glob
import json
import os
import pathlib
import re
import shutil
import urllib.parse

import augeas

from plinth import action_utils, utils
from plinth.actions import privileged, secret_str

openidc_config_path = pathlib.Path(
    '/etc/apache2/conf-available/freedombox-openidc.conf')
metadata_dir_path = pathlib.Path(
    '/var/cache/apache2/mod_auth_openidc/metadata/')


def _get_sort_key_of_version(version):
    """Return the sort key for a given version string.

    Simple implementation hoping that PHP Apache module version numbers will be
    simple.

    """
    parts = []
    for part in version.split('.'):
        try:
            parts.append(int(part))
        except ValueError:
            parts.append(part)

    return parts


def _sort_versions(versions):
    """Return a list of sorted version strings."""
    return sorted(versions, key=_get_sort_key_of_version, reverse=True)


def _disable_mod_php(webserver):
    """Disable all mod_php versions.

    Idempotent and harmless if all or no PHP modules are identified.
    Problematic if only some modules are found.

    """
    paths = glob.glob('/etc/apache2/mods-available/php*.conf')
    versions = []
    for path in paths:
        match = re.search(r'\/php(.*)\.conf$', path)
        if match:
            versions.append(match[1])

    versions = _sort_versions(versions)

    for version in versions:
        webserver.disable('php' + version, kind='module')


@privileged
def setup(old_version: int):
    """Setup Apache configuration."""
    # Regenerate the snakeoil self-signed SSL certificate. This is so that
    # FreedomBox images don't all have the same certificate. When FreedomBox
    # package is installed via apt, don't regenerate. When upgrading to newer
    # version of Apache FreedomBox app and setting up for the first time don't
    # regenerate.
    if action_utils.is_disk_image() and old_version == 0:
        action_utils.run([
            'make-ssl-cert', 'generate-default-snakeoil', '--force-overwrite'
        ], check=True)
    # In case the certificate has been removed after ssl-cert is installed
    # on a fresh Debian machine.
    elif not os.path.exists('/etc/ssl/certs/ssl-cert-snakeoil.pem'):
        action_utils.run(['make-ssl-cert', 'generate-default-snakeoil'],
                         check=True)

    _setup_oidc_config()

    with action_utils.WebserverChange() as webserver:
        # Disable mod_php as we have switched to mod_fcgi + php-fpm. Disable
        # before switching away from mpm_prefork otherwise switching fails due
        # dependency.
        _disable_mod_php(webserver)

        # set the prefork worker model
        webserver.disable('mpm_worker', kind='module')
        webserver.disable('mpm_prefork', kind='module')
        webserver.enable('mpm_event', kind='module')

        # enable miscellaneous modules.
        webserver.enable('proxy', kind='module')
        webserver.enable('proxy_http', kind='module')
        webserver.enable('proxy_fcgi', kind='module')
        webserver.enable('proxy_html', kind='module')
        webserver.enable('rewrite', kind='module')
        webserver.enable('macro', kind='module')
        webserver.enable('expires', kind='module')

        # Disable logging into files, use FreedomBox configured systemd logging
        webserver.disable('other-vhosts-access-log', kind='config')

        # Disable /server-status page to avoid leaking private info.
        webserver.disable('status', kind='module')

        # Enable HTTP/2 protocol
        webserver.enable('http2', kind='module')

        # Enable shared object cache needed for OSCP stapling. Needed by
        # mod_ssl.
        webserver.enable('socache_shmcb', kind='module')

        # switch to mod_ssl from mod_gnutls
        webserver.disable('gnutls', kind='module')
        webserver.enable('ssl', kind='module')

        # enable mod_alias for RedirectMatch
        webserver.enable('alias', kind='module')

        # enable mod_headers for HSTS
        webserver.enable('headers', kind='module')

        # Various modules for authentication/authorization
        webserver.enable('auth_openidc', kind='module')
        webserver.enable('authnz_ldap', kind='module')
        webserver.disable('auth_pubtkt', kind='module')

        # enable some critical modules to avoid restart while installing
        # FreedomBox applications.
        webserver.disable('cgi', kind='module')  # For process MPMs
        webserver.enable('cgid', kind='module')  # For threaded MPMs
        webserver.enable('proxy_uwsgi', kind='module')
        webserver.enable('proxy_wstunnel', kind='module')

        # enable configuration for PHP-FPM
        webserver.enable('php-fpm-freedombox', kind='config')

        # enable users to share files uploaded to ~/public_html
        webserver.enable('userdir', kind='module')

        # enable WebDAV protocol. Used by feather wiki and potentially by other
        # apps and file sharing.
        webserver.enable('dav', kind='module')
        webserver.enable('dav_fs', kind='module')

        # setup freedombox configuration
        webserver.enable('10-freedombox', kind='config')
        webserver.enable('freedombox', kind='config')
        webserver.enable('freedombox-tls', kind='config')
        webserver.enable('freedombox-openidc.conf', kind='config')

        # enable serving Debian javascript libraries
        webserver.enable('javascript-common', kind='config')

        # default sites
        webserver.disable('000-default', kind='site')
        webserver.disable('default-tls', kind='site')
        webserver.disable('default-ssl', kind='site')
        webserver.disable('plinth', kind='site')
        webserver.disable('plinth-ssl', kind='site')
        webserver.enable('freedombox-default', kind='site')


def _load_augeas():
    """Initialize augeas for this app's configuration file."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Httpd', str(openidc_config_path))
    aug.set('/augeas/context', '/files' + str(openidc_config_path))
    aug.load()

    return aug


def _get_mod_openidc_passphrase() -> str:
    """Read existing mod-auth-openidc passphase.

    Instead of generating a new passphrase every time, use existing one. If the
    passphrase changes, all the existing sessions will be logged out and users
    will have login to apps again.
    """
    aug = _load_augeas()
    for directive in aug.match('*/directive'):
        if aug.get(directive) == 'OIDCCryptoPassphrase':
            return aug.get(directive + '/arg')

    # Does not exist already, generate new
    return utils.generate_password(size=64)


@privileged
def setup_oidc_client(netloc: str, client_id: str, client_secret: secret_str):
    """Setup client ID and secret for provided domain.

    netloc is hostname or IP address along with port as parsed by
    urllib.parse.urlparse() method from a URL.
    """
    issuer = f'{netloc}/freedombox/o'
    issuer_quoted = urllib.parse.quote_plus(issuer)
    client_path = metadata_dir_path / f'{issuer_quoted}.client'
    if client_path.exists():
        try:
            current_data = json.loads(client_path.read_text())
            if (current_data['client_id'] == client_id
                    and current_data['client_secret'] == client_secret):
                return
        except Exception:
            pass

    client_configuration = {
        'client_id': client_id,
        'client_secret': client_secret
    }
    previous_umask = os.umask(0o077)
    try:
        client_path.write_text(json.dumps(client_configuration))
    finally:
        os.umask(previous_umask)

    shutil.chown(client_path, 'www-data', 'www-data')


def _setup_oidc_config():
    """Setup Apache as a OpenID Connect Relying Party.

    Ensure that auth_openidc module's metadata directory is created. It will be
    used to store provider-specific configuration. Since FreedomBox will be
    configured with multiple domains and some of them may not be accessible due
    to the access method, we need to configure a separate IDP for each domain.
    This is also because auth_openidc does not allow IDP configuration with
    relative URLs.

    Keep the metadata directory and configuration file unreadable by non-admin
    users since they contain module's crypto secret and OIDC client secret.

    # Session management

    When apps like Syncthing are protected with mod_auth_openidc, there a
    session maintained using server-side session storage and a cookie on the
    client side. This session is different from the session managed by the
    OpenID Connect Provider (FreedomBox web interface). As long as this session
    is valid, no further authentication mechanisms are triggered.

    When the session expires, if the request is a GET request (due to page
    reload), the browser is redirected to OP, a fresh token is created, and the
    page is loaded. However, for POST requests, 401 error is returned and if
    the application is unaware, it won't do much about it. So, this
    necessitates keeping the session timeout value high.

    When logout is performed on FreedomBox web interface, mod_auth_openidc
    cookie is also removed and logout feels instantaneous. However, this won't
    work for applications not using mod_auth_openidc and for applications
    hosted on other domains. A better way to do this is to implement OpenID
    Connect's Back-Channel Logout or using OpenID Connect Session Management.

    For more about session management see:
    https://github.com/OpenIDC/mod_auth_openidc/wiki/Sessions-and-Timeouts and
    https://github.com/OpenIDC/mod_auth_openidc/wiki/Session-Management-Settings
    """
    metadata_dir_path.parent.mkdir(mode=0o700, parents=True, exist_ok=True)
    metadata_dir_path.mkdir(mode=0o700, exist_ok=True)
    shutil.chown(metadata_dir_path.parent, 'www-data', 'www-data')
    shutil.chown(metadata_dir_path, 'www-data', 'www-data')

    # XXX: Default cache type is 'shm' or shared memory. This is lost when
    # Apache is restarted and users/apps will have to reauthenticate. Improve
    # this by using file (in tmpfs), redis, or memache caches.

    passphrase = _get_mod_openidc_passphrase()
    configuration = f'''##
## OpenID Connect related configuration
##
<IfModule mod_auth_openidc.c>
    OIDCCryptoPassphrase {passphrase}
    OIDCMetadataDir {str(metadata_dir_path)}
    # Use relative URL to redirect to the same origin as the resource
    OIDCDiscoverURL /freedombox/apache/discover-idp/
    OIDCSSLValidateServer Off
    OIDCProviderMetadataRefreshInterval 86400

    # Expire the mod_auth_openidc session (not the OpenID Conneect Provider
    # session) after 10 hours of idle with a maximum session duration equal to
    # the expiry time of the ID token (10 hours). This allows applications such
    # as FeatherWiki to have long editing sessions before save.
    OIDCSessionInactivityTimeout 36000
    OIDCSessionMaxDuration 0

    # Use relative URL to return to the original domain
    OIDCRedirectURI /apache/oidc/callback
    OIDCRemoteUserClaim sub

    # The redirect URI must always be under a location protected by
    # mod_openidc.
    <Location /apache>
        AuthType openid-connect
        # Checking audience is not necessary, but we need to check some claim.
        Require claim aud:apache
    </Location>
</IfModule>
'''
    previous_umask = os.umask(0o077)
    try:
        openidc_config_path.write_text(configuration)
    finally:
        os.umask(previous_umask)


# TODO: Check that the (name, kind) is a managed by FreedomBox before
# performing operation.
@privileged
def enable(name: str, kind: str):
    """Enable an Apache site/config/module."""
    _assert_kind(kind)
    action_utils.webserver_enable(name, kind)


@privileged
def disable(name: str, kind: str):
    """Disable an Apache site/config/module."""
    _assert_kind(kind)
    action_utils.webserver_disable(name, kind)


def _assert_kind(kind: str):
    """Raise and exception if kind parameter has an unexpected value."""
    if kind not in ('site', 'config', 'module'):
        raise ValueError('Invalid value for parameter kind')


@privileged
def link_root(domain: str, name: str):
    """Link the Apache site root configuration to app configuration."""
    if '/' in domain or '/' in name:
        raise ValueError('Invalid domain or name')

    target_config = f'{name}.conf'

    include_root = pathlib.Path('/etc/apache2/includes/')
    config = include_root / f'{domain}-include-freedombox.conf'
    config.unlink(missing_ok=True)
    config.symlink_to(target_config)
    action_utils.service_reload('apache2')


@privileged
def unlink_root(domain: str):
    """Unlink the Apache site root configuration from app configuration."""
    if '/' in domain:
        raise ValueError('Invalid domain')

    include_root = pathlib.Path('/etc/apache2/includes/')
    config = include_root / f'{domain}-include-freedombox.conf'
    if not config.is_symlink():
        return  # Does not exist or not a symlink

    config.unlink()
    action_utils.service_reload('apache2')


@privileged
def domain_setup(domain: str):
    """Add site specific configuration for a domain."""
    if '/' in domain:
        raise ValueError('Invalid domain')

    path = pathlib.Path('/etc/apache2/sites-available/')
    path = path / (domain + '.conf')
    configuration = 'Use FreedomBoxTLSSiteMacro {domain}\n'
    if path.is_file():
        return  # File already exists. Assume it to be correct one.

    path.write_text(configuration.format(domain=domain))

    with action_utils.WebserverChange() as webserver:
        webserver.enable('freedombox-tls-site-macro', kind='config')
        webserver.enable(domain, kind='site')


@privileged
def domain_remove(domain: str):
    """Remove site specific configuration for a domain."""
    if '/' in domain:
        raise ValueError('Invalid domain')

    with action_utils.WebserverChange() as webserver:
        webserver.disable(domain, kind='site')

    path = pathlib.Path('/etc/apache2/sites-available/')
    path = path / (domain + '.conf')
    path.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for name services.
"""

import logging

from django.contrib import messages
from django.shortcuts import redirect
from django.urls import reverse_lazy
from django.utils.translation import gettext_lazy as _
from django.views.generic import TemplateView
from django.views.generic.edit import FormView

from plinth.modules import names
from plinth.signals import domain_added, domain_removed
from plinth.views import AppView

from . import components, privileged, resolved
from .forms import DomainAddForm, HostnameForm, NamesConfigurationForm

logger = logging.getLogger(__name__)


class NamesAppView(AppView):
    """Show names app main page."""

    app_id = 'names'
    template_name = 'names.html'
    prefix = 'names'
    form_class = NamesConfigurationForm

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        if names.is_resolved_installed():
            initial.update(privileged.get_resolved_configuration())

        return initial

    def get_context_data(self, *args, **kwargs):
        """Add additional context data for template."""
        context = super().get_context_data(*args, **kwargs)
        context['status'] = get_status()
        context['resolved_installed'] = names.is_resolved_installed()
        if context['resolved_installed']:
            try:
                context['resolved_status'] = resolved.get_status()
            except Exception as exception:
                context['resolved_status_error'] = exception

        return context

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_data = form.initial
        form_data = form.cleaned_data

        changes = {}
        if old_data['dns_over_tls'] != form_data['dns_over_tls']:
            changes['dns_over_tls'] = form_data['dns_over_tls']

        if old_data['dnssec'] != form_data['dnssec']:
            changes['dnssec'] = form_data['dnssec']

        if changes:
            privileged.set_resolved_configuration(**changes)
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)


class HostnameView(FormView):
    """View to update system's hostname."""
    template_name = 'form.html'
    form_class = HostnameForm
    prefix = 'hostname'
    success_url = reverse_lazy('names:index')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Set Hostname')
        return context

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial['hostname'] = names.get_hostname()
        return initial

    def form_valid(self, form):
        """Apply the form changes."""
        if form.initial['hostname'] != form.cleaned_data['hostname']:
            try:
                names.set_hostname(form.cleaned_data['hostname'])
                messages.success(self.request, _('Configuration updated'))
            except Exception as exception:
                messages.error(
                    self.request,
                    _('Error setting hostname: {exception}').format(
                        exception=exception))

        return super().form_valid(form)


class DomainAddView(FormView):
    """View to update system's static domain name."""
    template_name = 'names-domain-add.html'
    form_class = DomainAddForm
    prefix = 'domain-add'
    success_url = reverse_lazy('names:index')

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['title'] = _('Add Domain Name')
        return context

    def form_valid(self, form):
        """Apply the form changes."""
        _domain_add(form.cleaned_data['domain_name'])
        messages.success(self.request, _('Configuration updated'))
        return super().form_valid(form)


class DomainDeleteView(TemplateView):
    """Confirm and delete a domain."""
    template_name = 'names-domain-delete.html'

    def get_context_data(self, **kwargs):
        """Return additional context data for rendering the template."""
        context = super().get_context_data(**kwargs)
        domain = self.kwargs['domain']
        context['domain'] = domain
        context['title'] = str(
            _('Delete Domain {domain}?')).format(domain=domain)
        return context

    def post(self, request, domain):
        """Delete a domain."""
        _domain_delete(domain)
        messages.success(request, _('Domain deleted.'))
        return redirect('names:index')


def get_status() -> dict[str, object]:
    """Get configured services per name."""
    domains = components.DomainName.list()
    used_domain_types = {domain.domain_type for domain in domains}
    unused_domain_types = [
        domain_type for domain_type in components.DomainType.list().values()
        if domain_type not in used_domain_types or domain_type.add_url
    ]

    return {'domains': domains, 'unused_domain_types': unused_domain_types}


def _domain_add(domain_name: str):
    """Add a static domain name."""
    # Domain name is not case sensitive, but Let's Encrypt certificate
    # paths use lower-case domain name.
    domain_name = domain_name.lower()

    logger.info('Adding domain name - %s', domain_name)
    names.domain_add(domain_name)

    domain_added.send_robust(sender='names', domain_type='domain-type-static',
                             name=domain_name, services='__all__')


def _domain_delete(domain_name: str):
    """Remove a static domain name."""
    # Domain name is not case sensitive, but Let's Encrypt certificate
    # paths use lower-case domain name.
    domain_name = domain_name.lower()

    logger.info('Removing domain name - %s', domain_name)
    names.domain_delete(domain_name)

    # Update domain registered with Name Services module.
    domain_removed.send_robust(sender='names',
                               domain_type='domain-type-static',
                               name=domain_name)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""App component for other apps to use handle Let's Encrypt certificates."""

import logging
import pathlib
import threading
from typing import ClassVar

from plinth import app
from plinth.modules.names.components import DomainName
from plinth.privileged import service as service_privileged

from . import privileged

logger = logging.getLogger(__name__)

_list_type = list


class LetsEncrypt(app.FollowerComponent):
    """Component to receive Let's Encrypt renewal hooks.

    Performs the following tasks:

    - Listen to certificate change events from the Let's Encrypt app. This
      could be obtain, re-obtain, renewal, revoke and delete events.

    - Filter out events after comparing domains list in certificate events to
      the list of domains the app (to which this component belongs) is
      interested in.

    - If requested, copy certificates from /etc/letsencrypt directory to app's
      directory with the needed permissions and ownership.

    - Query status of copied certificates. Status could be up-to-date,
      self-signed, outdated (/etc/letsencrypt has more recent certificate),
      expired or a test certificate.

    - Optionally restart a bunch of daemons after interesting certificate event
      occurs.

    """

    _all: ClassVar[dict[str, 'LetsEncrypt']] = {}

    def __init__(self, component_id, domains=None, daemons=None,
                 should_copy_certificates=False, private_key_path=None,
                 certificate_path=None, user_owner=None, group_owner=None,
                 managing_app=None, reload_daemons=False):
        """Initialize the Let's Encrypt component.

        component_id should be a unique ID across all components of an app and
        across all components.

        domains is a list of domains that the app is configured for. Any
        certificate operation that is not happening on these domains is ignored
        by this component.

        domains can be the string '*' in which case it means that the app is
        relevant for (or is interested in) all domains configured in the
        system.

        domains can be a callable instead of a list. In this case, the callable
        is called every time an operation needs to be checked against list of
        domains for the app. The callable must return a list of domains that
        the app is configured for.

        daemons is a list of systemd units that need to be restarted (actually
        try-restarted, i.e., only if they are already running) when a relevant
        certificate operation happens.

        should_copy_certificates indicates whether the app needs the Let's
        Encrypt certificates in the directory /etc/letsencrypt to be copied to
        a different location with different permissions to be able to use them.
        If true, private_key_path, certificate_path, user_owner, group_owner
        and managing_app properties must be specified. Otherwise, ValueError is
        raised.

        private_key_path is the string containing the file path to which a
        private key must be copied. It may contain the sub-string '{domain}'
        which will be replaced with the name of the domain for which this
        certificate is being copied. Any non-existing parent directories are
        created with 'root:root' ownership and '0o755' file permissions. The
        permissions for the file will be '0o600' and ownership will be as
        specified in parameters user_owner and group_owner. If both
        private_key_path and certificate_path refer to the same path, the
        target file is created by appending the private key and certificate.

        certificate_path is the string containing the file path to which a
        certificate must be copied. It may contain the sub-string '{domain}'
        which will be replaced with the name of the domain for which this
        certificate is being copied. Any non-existing parent directories are
        created with 'root:root' ownership and '0o755' file permissions. The
        permissions for the file will be '0o644' and ownership will be as
        specified in parameters user_owner and group_owner. If both
        private_key_path and certificate_path refer to the same path, the
        target file is created by appending the private key and certificate.

        user_owner specifies the user who should own the copied certificates.
        This is typically the unprivileged user account that runs the daemon.

        group_owner specifies the group who should own the copied certificates.
        This is typically the unprivileged group account that runs the daemon.

        managing_app is the name of the app that contains this component. This
        is used by the superuser action to verify that this app is allowed to
        copy certificates to the target location. The app must then define a
        property 'managing_paths' which contains a list of pathlib.Path()
        objects to which the app is allowed to write certificates and other
        files to.

        reload_daemons is boolean to indicate if the daemons should be reloaded
        instead of restarted (which is default).
        """
        if should_copy_certificates:
            if (not private_key_path or not certificate_path or not user_owner
                    or not group_owner or not managing_app):
                raise ValueError('Not enough values for copying certificates')

        super().__init__(component_id)
        self._domains = domains
        self.daemons = daemons
        self.reload_daemons = reload_daemons
        self.should_copy_certificates = should_copy_certificates
        self.private_key_path = private_key_path
        self.certificate_path = certificate_path
        self.user_owner = user_owner
        self.group_owner = group_owner
        self.managing_app = managing_app

        self._all[component_id] = self

    @property
    def domains(self) -> list[str] | str:
        """Return a list of domains this component's app is interested in."""
        if callable(self._domains):
            return self._domains()

        return self._domains

    @classmethod
    def list(cls):
        """Return a list of all Let's Encrypt components."""
        return cls._all.values()

    def setup_certificates(
            self, app_domains: str | _list_type[str] | None = None) -> None:
        """Setup app certificates for all interested domains.

        For every domain, a certificate is copied. If a valid certificate is
        not available in Let's Encrypt, a self-signed snakeoil certificate is
        used. Each daemon is restarted if it is running.

        app_domains is the list of domains for which certificates must be
        copied. If it is not provided, the component's list of domains (which
        may be acquired by a callable) is used.
        """
        if not app_domains:
            app_domains = self.domains

        if app_domains == '*':
            # Setup for all domains and not just ones that LE can obtain
            # certificate for. This allows the domains that can't have LE
            # certificate to work with self-signed certificates.
            app_domains = DomainName.list_names()

        domains, status = self._get_letsencrypt_domains()

        if self.should_copy_certificates:
            for domain in app_domains:
                if domain in domains:
                    lineage = status['domains'][domain]['lineage']
                    self._copy_letsencrypt_certificates([domain], lineage)
                else:
                    self._copy_self_signed_certificates([domain])

        for daemon in self.daemons:
            if self.reload_daemons:
                service_privileged.try_reload_or_restart(daemon)
            else:
                service_privileged.try_restart(daemon)

    def get_status(self):
        """Return the status of certificates for all interested domains.

        A dictionary is returned containing a key for each domain and the value
        shall be a simple string with state of certificate for this app.
        Possible values are 'outdated-copy', 'self-signed', 'unknown', 'valid',
        'revoked', 'expired' and 'test_cert'.

        """
        app_domains = self.domains
        domains, le_status = self._get_letsencrypt_domains()

        final_status = {}
        for domain in app_domains:
            if domain in domains:
                status = le_status['domains'][domain]['validity']
                if self.should_copy_certificates:
                    if not self._compare_certificate(
                            domain, le_status['domains'][domain]['lineage']):
                        status = 'outdated-copy'
            else:
                status = 'self-signed'

            final_status[domain] = status

        return final_status

    def on_certificate_obtained(self, domains, lineage):
        """Handle event when a certificate is obtained.

        Filter any domains that are interesting for the app that this component
        is part of. For all interesting domains, copy the certificates if
        requested and then restart the daemons.

        """
        interested_domains = self._get_interested_domains(domains)
        if not interested_domains:
            return

        if self.should_copy_certificates:
            self._copy_letsencrypt_certificates(interested_domains, lineage)

        for daemon in self.daemons:
            if self.reload_daemons:
                service_privileged.try_reload_or_restart(daemon)
            else:
                service_privileged.try_restart(daemon)

    def on_certificate_renewed(self, domains, lineage):
        """Handle event when a certificate is renewed.

        This is typically called when a Cron job triggers certbot renewal
        process which in turn triggers configured renewal hooks after
        successful renewal. It is also called when a valid certificate is
        re-obtained.

        To trigger a renewal event for testing purposes, run `certbot renew
        --force-renewal`.

        For now, performs the same operations as obtaining a fresh certificate.

        """
        return self.on_certificate_obtained(domains, lineage)

    def on_certificate_revoked(self, domains, lineage):
        """Handle event when a certificate is revoked.

        Filter any domains that are interesting for the app that this component
        is part of. For all interesting domains, copy self-signed snakeoil
        certificates if requested and then restart the daemons.

        """
        interested_domains = self._get_interested_domains(domains)
        if not interested_domains:
            return

        if self.should_copy_certificates:
            self._copy_self_signed_certificates(interested_domains)

        for daemon in self.daemons:
            if self.reload_daemons:
                service_privileged.try_reload_or_restart(daemon)
            else:
                service_privileged.try_restart(daemon)

    def on_certificate_deleted(self, domains, lineage):
        """Handle event when a certificate is deleted.

        For now, performs the same operations as obtaining a fresh certificate.

        """
        return self.on_certificate_revoked(domains, lineage)

    @staticmethod
    def _get_letsencrypt_domains():
        """Return the list of domains with LE certificates."""
        from plinth.modules import letsencrypt
        status = letsencrypt.get_status()
        domains = {
            domain
            for domain, domain_status in status['domains'].items()
            if domain_status
        }
        return domains, status

    def _get_interested_domains(self, domains):
        """Return set of domains the component is interested in."""
        app_domains = self.domains
        if not app_domains:
            return set()

        if app_domains == '*':
            return set(domains)

        return set(domains).intersection(set(app_domains))

    def _copy_self_signed_certificates(self, domains):
        """Copy a self-signed certificate for all domains.

        Copy Apache's snake-oil certificate into daemon's directory. The
        self-signed certificate may not really work. It is merely to prevent
        the server from failing to startup because the files are missing.

        """
        source_private_key_path = '/etc/ssl/private/ssl-cert-snakeoil.key'
        source_certificate_path = '/etc/ssl/certs/ssl-cert-snakeoil.pem'
        self._copy_certificates(domains, source_private_key_path,
                                source_certificate_path)

    def _copy_letsencrypt_certificates(self, domains, lineage):
        """Copy a valid Let's Encrypt certificate for given domains."""
        source_private_key_path = pathlib.Path(lineage) / 'privkey.pem'
        source_certificate_path = pathlib.Path(lineage) / 'fullchain.pem'
        self._copy_certificates(domains, source_private_key_path,
                                source_certificate_path)

    def _copy_certificates(self, domains, source_private_key_path,
                           source_certificate_path):
        """Copy certificate for all domains in the certificate.

        The list of domains passed should be subset of domains listed in the
        certificate.

        """
        if '{domain}' in self.private_key_path or \
           '{domain}' in self.certificate_path:
            for domain in domains:
                private_key_path = self.private_key_path.format(domain=domain)
                certificate_path = self.certificate_path.format(domain=domain)
                self._copy_certificate(source_private_key_path,
                                       source_certificate_path,
                                       private_key_path, certificate_path)
        else:
            self._copy_certificate(source_private_key_path,
                                   source_certificate_path,
                                   self.private_key_path,
                                   self.certificate_path)

    def _copy_certificate(self, source_private_key_path,
                          source_certificate_path, private_key_path,
                          certificate_path):
        """Copy certificate for a single domain."""
        privileged.copy_certificate(self.managing_app,
                                    str(source_private_key_path),
                                    str(source_certificate_path),
                                    private_key_path, certificate_path,
                                    self.user_owner, self.group_owner)

    def _compare_certificate(self, domain, lineage):
        """Compare LE certificate with app certificate."""
        source_private_key_path = pathlib.Path(lineage) / 'privkey.pem'
        source_certificate_path = pathlib.Path(lineage) / 'fullchain.pem'
        private_key_path = self.private_key_path.format(domain=domain)
        certificate_path = self.certificate_path.format(domain=domain)
        return privileged.compare_certificate(self.managing_app,
                                              str(source_private_key_path),
                                              str(source_certificate_path),
                                              private_key_path,
                                              certificate_path)


def on_certificate_event(event, domains, lineage):
    """Start a new thread to handle a LE certificate event.

    Run in a new thread because:

    - We don't want to block the thread running Glib main loop when called from
      dbus handler in case of renewal.

    - We don't want to delay the a certificate operation with copying
      certificates to all apps.

    - We don't want to cause a page load error when restarting Apache due to
      certificate changes.

    """
    threading.Thread(target=on_certificate_event_sync,
                     args=(event, domains, lineage)).start()


def on_certificate_event_sync(event, domains, lineage):
    """Trigger certificate event hooks across all apps."""
    if isinstance(domains, str):
        domains = domains.split()

    if not lineage:
        # XXX: A better to find lineage is by reading the renewal/{domain}.conf
        lineage = str(pathlib.Path('/etc/letsencrypt/live/') / domains[0])

    assert event in ('obtained', 'renewed', 'revoked', 'deleted')

    for component in LetsEncrypt.list():
        if component.app and component.app.needs_setup():
            continue

        logger.info('Handling certificate event for %s: %s, %s, %s',
                    component.component_id, event, domains, lineage)

        try:
            getattr(component, 'on_certificate_' + event)(domains, lineage)
        except Exception as exception:
            logger.exception(
                'Error executing certificate hook for %s: %s, %s, %s: %s',
                component.component_id, event, domains, lineage, exception)

    if event in ('obtained', 'renewed'):
        from plinth.modules import letsencrypt
        letsencrypt.certificate_set_last_seen_modified_time(lineage)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Remote and local Borg backup repositories."""

import abc
import datetime
import logging
import os
import subprocess
from uuid import uuid1

from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.utils import format_lazy

from . import (_backup_handler, api, copy_ssh_client_public_key, errors,
               generate_ssh_client_auth_key, get_known_hosts_path,
               get_ssh_client_auth_key_paths, privileged, raise_ssh_error,
               restore_archive_handler, split_path, store)
from .schedule import Schedule

logger = logging.getLogger(__name__)


class BaseBorgRepository(abc.ABC):
    """Base class for all kinds of Borg repositories."""
    flags: dict[str, bool] = {}
    is_mounted = True
    known_credentials: list[str] = []

    def __init__(self, path, credentials=None, uuid=None, schedule=None,
                 **kwargs):
        """Instantiate a new repository."""
        self._path = path
        self.credentials = credentials or {}
        self.uuid = uuid or str(uuid1())
        self.kwargs = kwargs
        schedule = schedule or {}
        schedule['repository_uuid'] = self.uuid
        self.schedule = Schedule(**schedule)

    @classmethod
    def load(cls, uuid):
        """Create instance by loading from database."""
        storage = dict(store.get(uuid))
        path = storage.pop('path')
        storage.pop('uuid')
        credentials = storage.setdefault('credentials', {})
        storage.pop('credentials')
        schedule = storage.setdefault('schedule', {})
        storage.pop('schedule')

        return cls(path, credentials, uuid, schedule, **storage)

    @property
    def name(self):
        """Return a display name for the repository."""
        return self._path

    @property
    def path(self):
        """Return the path of the repository."""
        return self._path

    @abc.abstractmethod
    def storage_type(self):
        """Return the storage type of repository."""
        raise NotImplementedError

    @staticmethod
    def is_usable():
        """Return whether the repository is ready to be used."""
        return True

    @property
    def borg_path(self):
        """Return the repository that the backups action script should use."""
        return self._path

    @staticmethod
    def prepare():
        """Prepare the repository for operations."""

    @staticmethod
    def cleanup():
        """Cleanup the repository after operations."""

    def get_info(self):
        """Return Borg information about a repository."""
        output = privileged.info(self.borg_path,
                                 self._get_encryption_passpharse())
        if output['encryption']['mode'] == 'none' and \
           self._get_encryption_data():
            raise errors.BorgUnencryptedRepository(
                _('Existing repository is not encrypted.'))

        return output

    def get_view_content(self):
        """Get archives with additional information as needed by the view."""
        repository = {
            'uuid': self.uuid,
            'name': self.name,
            'storage_type': self.storage_type,
            'is_encrypted': bool(self._get_encryption_data()),
            'flags': self.flags,
            'error': None,
        }
        try:
            repository['mounted'] = self.is_mounted
            if repository['mounted']:
                repository['archives'] = self.list_archives()
        except (errors.BorgError, Exception) as err:
            repository['error'] = str(err)

        return repository

    def remove(self):
        """Remove a borg repository"""

    def list_archives(self):
        """Return list of archives in this repository."""
        archives = privileged.list_repo(
            self.borg_path, self._get_encryption_passpharse())['archives']
        for archive in archives:
            archive['time'] = datetime.datetime.strptime(
                archive['time'], '%Y-%m-%dT%H:%M:%S.%f')
            archive['start'] = datetime.datetime.strptime(
                archive['start'], '%Y-%m-%dT%H:%M:%S.%f')

        return sorted(archives, key=lambda archive: archive['start'],
                      reverse=True)

    def create_archive(self, archive_name, app_ids, archive_comment=None):
        """Create a new archive in this repository with given name."""
        archive_path = self._get_archive_path(archive_name)
        passphrase = self.credentials.get('encryption_passphrase', None)
        api.backup_apps(_backup_handler, path=archive_path, app_ids=app_ids,
                        encryption_passphrase=passphrase,
                        archive_comment=archive_comment)

    def delete_archive(self, archive_name):
        """Delete an archive with given name from this repository."""
        archive_path = self._get_archive_path(archive_name)
        privileged.delete_archive(archive_path,
                                  self._get_encryption_passpharse())

    def migrate_credentials(self) -> None:
        """Migrate any credentials."""

    def initialize(self):
        """Initialize / create a borg repository."""
        encryption = 'none'
        if 'encryption_passphrase' in self.credentials and \
           self.credentials['encryption_passphrase']:
            encryption = 'repokey'

        try:
            privileged.init(self.borg_path, encryption,
                            self._get_encryption_passpharse())
        except errors.BorgRepositoryExists:
            pass

        self.get_info()  # If password is incorrect raise an error early.

    def _get_encryption_data(self):
        """Return additional dictionary data to send to backups call."""
        passphrase = self.credentials.get('encryption_passphrase', None)
        if passphrase:
            return {'encryption_passphrase': passphrase}

        return {}

    def _get_encryption_passpharse(self):
        """Return encryption passphrase or raise an exception."""
        for key in self.credentials.keys():
            if key not in self.known_credentials:
                raise ValueError('Unknown credentials entry: %s' % key)

        return self.credentials.get('encryption_passphrase', None)

    def get_download_stream(self, archive_name):
        """Return an stream of .tar.gz binary data for a backup archive."""
        reader = privileged.export_tar(self._get_archive_path(archive_name),
                                       self._get_encryption_passpharse(),
                                       _raw_output=True)
        return reader

    def _get_archive_path(self, archive_name):
        """Return full borg path for an archive."""
        return '::'.join([self.borg_path, archive_name])

    def get_archive(self, name):
        """Return a specific archive from this repository with given name."""
        for archive in self.list_archives():
            if archive['name'] == name:
                return archive

        return None

    def generate_archive_name(self):
        """Return a name to create a backup archive with using time."""
        return datetime.datetime.now().astimezone().replace(
            microsecond=0).isoformat()

    def get_archive_apps(self, archive_name):
        """Get list of apps included in an archive."""
        archive_path = self._get_archive_path(archive_name)
        return privileged.get_archive_apps(archive_path,
                                           self._get_encryption_passpharse())

    def restore_archive(self, archive_name, app_ids=None):
        """Restore an archive from this repository to the system."""
        archive_path = self._get_archive_path(archive_name)
        passphrase = self.credentials.get('encryption_passphrase', None)
        api.restore_apps(restore_archive_handler, app_ids=app_ids,
                         create_subvolume=False, backup_file=archive_path,
                         encryption_passphrase=passphrase)

    def _get_storage_format(self):
        """Return a dict representing the repository."""
        storage = {
            'path': self._path,
            'storage_type': self.storage_type,
            'added_by_module': 'backups',
            'credentials': self.credentials,
            'schedule': self.schedule.get_storage_format()
        }
        if self.uuid:
            storage['uuid'] = self.uuid

        return storage

    def save(self):
        """Save the repository in store (kvstore)."""
        storage = self._get_storage_format()
        self.uuid = store.update_or_add(storage)


class RootBorgRepository(BaseBorgRepository):
    """Borg repository on the root filesystem."""
    UUID = 'root'
    PATH = '/var/lib/freedombox/borgbackup'

    storage_type = 'root'  # type: ignore
    name = format_lazy(_('{box_name} storage'), box_name=_(cfg.box_name))
    borg_path = PATH
    sort_order = 10
    is_mounted = True

    def __init__(self, path=None, credentials=None, uuid=None, schedule=None,
                 **kwargs):
        """Initialize the repository object."""
        super().__init__(self.PATH, credentials, self.UUID, schedule, **kwargs)


class BorgRepository(BaseBorgRepository):
    """General Borg repository implementation."""
    known_credentials = ['encryption_passphrase']
    storage_type = 'disk'  # type: ignore
    sort_order = 20
    flags = {'removable': True}

    @property
    def name(self):
        # TODO Use disk label as the name
        # Also, name isn't being used yet
        return self._path

    def remove(self):
        """Remove a repository from the kvstore."""
        store.delete(self.uuid)


class SshBorgRepository(BaseBorgRepository):
    """Borg repository that is accessed via SSH"""
    SSHFS_MOUNTPOINT = '/media/'
    known_credentials = [
        'ssh_keyfile', 'ssh_password', 'encryption_passphrase'
    ]
    storage_type = 'ssh'  # type: ignore
    sort_order = 30
    flags = {'removable': True, 'mountable': True}

    def __init__(self, path, credentials=None, uuid=None, schedule=None,
                 verified=None, **kwargs):
        """Instantiate a new repository."""
        super().__init__(path, credentials, uuid, schedule, **kwargs)
        self.verified = verified or False

    def _get_storage_format(self):
        """Return a dict representing the repository."""
        storage = super()._get_storage_format()
        storage['verified'] = self.verified
        return storage

    def is_usable(self):
        """Return whether repository is usable."""
        return self.verified

    @property
    def borg_path(self):
        """Return the path to use for backups actions.

        This is the mount point for the remote SSH repositories.

        """
        return self._mountpoint

    def prepare(self):
        """Prepare the repository for operations by mounting."""
        if not self.is_usable():
            raise errors.SshfsError('Remote host not verified')

        self._umount_ignore_errors()  # In case the connection is stale.
        self.mount()

    def cleanup(self):
        """Cleanup the repository after operations by unmounting."""
        self._umount_ignore_errors()

    @property
    def hostname(self) -> str:
        """Return hostname from the remote path."""
        _, hostname, _ = split_path(self._path)
        return hostname.split('%')[0]  # XXX: Likely incorrect to split

    @property
    def username(self) -> str:
        """Return username from the remote path."""
        username, _, _ = split_path(self._path)
        return username

    @property
    def ssh_password(self) -> str | None:
        """Return SSH password if it is stored, otherwise None."""
        return self.credentials.get('ssh_password')

    @property
    def ssh_keyfile(self) -> str | None:
        """Return path to SSH client key if stored, otherwise None."""
        return self.credentials.get('ssh_keyfile')

    @property
    def _mountpoint(self):
        """Return the local mount point where repository is to be mounted."""
        return os.path.join(self.SSHFS_MOUNTPOINT, self.uuid)

    @property
    def is_mounted(self):
        """Return whether remote path is mounted locally."""
        return privileged.is_mounted(self._mountpoint)

    def migrate_credentials(self) -> None:
        """Add SSH keyfile credential and delete stored password."""
        if not self.ssh_password:
            return

        pubkey_path, keyfile_path = get_ssh_client_auth_key_paths()
        generate_ssh_client_auth_key()
        copy_ssh_client_public_key(str(pubkey_path), self.hostname,
                                   self.username, self.ssh_password)
        self.credentials['ssh_keyfile'] = str(keyfile_path)
        self.credentials.pop('ssh_password', None)
        self.save()

    def initialize(self):
        """Initialize the repository after mounting the target directory."""
        self.migrate_credentials()
        self._ensure_remote_directory()
        self.mount()
        super().initialize()

    def mount(self):
        """Mount the remote path locally using sshfs."""
        if self.is_mounted:
            return

        self.migrate_credentials()
        known_hosts_path = get_known_hosts_path()
        privileged.mount(self._mountpoint, self._path,
                         ssh_keyfile=self.credentials['ssh_keyfile'],
                         user_known_hosts_file=str(known_hosts_path))

    def umount(self):
        """Unmount the remote path that was mounted locally using sshfs."""
        if not self.is_mounted:
            return

        privileged.umount(self._mountpoint)

    def _umount_ignore_errors(self):
        """Run unmount operation and ignore any exceptions thrown."""
        try:
            self.umount()
        except Exception as exception:
            logger.warning('Unable to unmount repository', exc_info=exception)

    def remove(self):
        """Remove a repository from the kvstore and delete its mountpoint."""
        self.umount()
        store.delete(self.uuid)
        try:
            if os.path.exists(self._mountpoint):
                try:
                    self.umount()
                except Exception:
                    pass
                if not os.listdir(self._mountpoint):
                    os.rmdir(self._mountpoint)
        except Exception as err:
            logger.error(err)

    def _ensure_remote_directory(self):
        """Create remote SSH directory if it does not exist."""
        username, hostname, dir_path = split_path(self.path)
        if dir_path == '':
            dir_path = '.'

        if dir_path[0] == '~':
            dir_path = '.' + dir_path[1:]

        # Ensure remote directory exists, check contents
        known_hosts_path = str(get_known_hosts_path())
        with raise_ssh_error():
            subprocess.run([
                'ssh', '-i',
                str(self.ssh_keyfile), '-o',
                f'UserKnownHostsFile={known_hosts_path}', '-o',
                'BatchMode=yes', f'{username}@{hostname}', 'mkdir', '-p',
                dir_path
            ], stdout=subprocess.PIPE, stderr=subprocess.PIPE, check=True)


def get_repositories():
    """Get all repositories of a given storage type."""
    repositories = []
    storages = store.get_storages()
    for uuid in storages:
        repositories.append(get_instance(uuid))

    if RootBorgRepository.UUID not in storages:
        repositories.append(get_instance(RootBorgRepository.UUID))

    return sorted(repositories, key=lambda x: x.sort_order)


def get_instance(uuid):
    """Create a local or SSH repository object instance."""
    if uuid == RootBorgRepository.UUID:
        try:
            return RootBorgRepository.load(uuid)
        except KeyError:
            return RootBorgRepository()

    storage = store.get(uuid)
    if storage['storage_type'] == 'ssh':
        return SshBorgRepository.load(uuid)

    return BorgRepository.load(uuid)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for networks.
"""

from django.utils.translation import gettext_lazy as _

tags = [_('Internet'), _('Wi-Fi'), _('Local network'), _('Topology')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for System Monitoring (cockpit-pcp) in ‘System’.
"""

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import menu
from plinth.daemon import Daemon
from plinth.modules.backups.components import BackupRestore
from plinth.package import Package, Packages

from . import manifest

name = _('Performance')

_description = [
    _('Performance app allows you to collect, store and view information '
      'about utilization of the hardware. This can give you basic insights '
      'into usage patterns and whether the hardware is overloaded by users '
      'and services.'),
    _('Performance metrics are collected by Performance Co-Pilot and can be '
      'viewed using the Cockpit app.'),
]


class PerformanceApp(app_module.App):
    """FreedomBox app for Performance."""

    app_id = 'performance'

    _version = 1

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Performance'), icon='fa-bar-chart',
                               description=_description,
                               manual_page='Performance',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-performance', info.name, info.icon,
                              info.tags, 'performance:index',
                              parent_url_name='system:administration',
                              order=40)
        self.add(menu_item)

        packages = Packages('packages-performance', [
            # For bookworm, we need cockpit-pcp (which depends on
            # cockpit-bridge and pcp). For trixie, cockpit-pcp is
            # replaced by cockpit-bridge, and we need to specify a
            # dependency on pcp. There is some issue with having a
            # virtual package specified, see #2475.
            Package('cockpit-pcp') | Package('cockpit-bridge'),
            'cockpit-bridge', 'pcp'
        ])
        self.add(packages)

        backup_restore = BackupRestore('backup-restore-performance',
                                       **manifest.backup)
        self.add(backup_restore)

        daemon_0 = Daemon('daemon-performance-0', 'pmcd.service',
                          listen_ports=None)
        self.add(daemon_0)

        daemon_1 = Daemon('daemon-performance-1', 'pmie.service',
                          listen_ports=None)
        self.add(daemon_1)

        daemon_2 = Daemon('daemon-performance-2', 'pmlogger.service',
                          listen_ports=None)
        self.add(daemon_2)

        daemon_3 = Daemon('daemon-performance-3', 'pmproxy.service',
                          listen_ports=None)
        self.add(daemon_3)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        if not old_version:
            self.enable()



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Utilities to configure Dovecot."""

import apt

from plinth.utils import Version


def is_version_24():
    """Return the currently installed version of Dovecot."""
    cache = apt.Cache()
    try:
        version = cache['dovecot-core'].installed.version
    except KeyError:
        return True

    return Version(version) >= Version('1:2.4')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for basic system configuration."""

import pathlib

import augeas
from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import frontpage, menu
from plinth.daemon import RelatedDaemon
from plinth.modules.apache import (get_users_with_website, user_of_uws_url,
                                   uws_url_of_user)
from plinth.package import Packages
from plinth.privileged import service as service_privileged

from . import manifest, privileged

_description = [
    _('Here you can set some general configuration options '
      'like webserver home page etc.')
]

ADVANCED_MODE_KEY = 'advanced_mode'


class ConfigApp(app_module.App):
    """FreedomBox app for basic system configuration."""

    app_id = 'config'

    _version = 6

    can_be_disabled = False

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()
        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True,
                               depends=['apache', 'firewall', 'names'
                                        ], name=_('General Configuration'),
                               icon='fa-cog', description=_description,
                               manual_page='Configure', tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-config', _('Configure'), info.icon,
                              info.tags, 'config:index',
                              parent_url_name='system:system', order=30)
        self.add(menu_item)

        packages = Packages('packages-config', ['zram-tools'])
        self.add(packages)

        daemon1 = RelatedDaemon('related-daemon-config1', 'systemd-journald')
        self.add(daemon1)

        daemon2 = RelatedDaemon('related-daemon-config2', 'rsyslog')
        self.add(daemon2)

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)

        if old_version <= 3:
            privileged.set_logging_mode('volatile')
        elif old_version == 4:
            privileged.set_logging_mode(privileged.get_logging_mode())

        if old_version <= 5:
            # Update the style of Apache home page redirection.
            home_page = get_home_page()
            if home_page == 'apache-default':
                change_home_page(home_page)

        # systemd-journald is socket activated, it may not be running and it
        # does not support reload.
        service_privileged.try_restart('systemd-journald')
        # rsyslog when enabled, is activated by syslog.socket (shipped by
        # systemd). See:
        # https://www.freedesktop.org/wiki/Software/systemd/syslog/ .
        service_privileged.disable('rsyslog')
        # Ensure that rsyslog is not started by something else as it is
        # installed by default on Debian systems.
        service_privileged.mask('rsyslog')


def home_page_url2scid(url: str | None):
    """Return the shortcut ID of the given home page url."""
    # url is None when the freedombox-apache-homepage configuration file exists
    # but is empty. In this case, no redirect is effective.
    if url in ('/index.html', None):
        return 'apache-default'

    if url in ('/plinth/', '/plinth', 'plinth', '/freedombox/', '/freedombox',
               'freedombox'):
        return 'freedombox'

    if url and url.startswith('/~'):
        return 'uws-{}'.format(user_of_uws_url(url))

    shortcuts = frontpage.Shortcut.list()
    for shortcut in shortcuts:
        if shortcut.url == url:
            return shortcut.component_id

    return None


def _home_page_scid2url(shortcut_id: str) -> str | None:
    """Return the url for the given home page shortcut ID."""
    url: str | None = '/freedombox/'
    if shortcut_id == 'freedombox':
        pass
    elif shortcut_id == 'apache-default':
        url = None
    elif shortcut_id.startswith('uws-'):
        user = shortcut_id[4:]
        if user in get_users_with_website():
            url = uws_url_of_user(user)
    else:
        shortcuts = frontpage.Shortcut.list()
        aux = [
            shortcut.url for shortcut in shortcuts
            if shortcut_id == shortcut.component_id
        ]
        if 1 == len(aux):
            url = aux[0]

    return url


def _get_home_page_url() -> str | None:
    """Get the default application for the domain."""
    conf_file = privileged.APACHE_HOMEPAGE_CONFIG
    if not pathlib.Path(conf_file).exists():
        return '/freedombox/'

    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Httpd/lens', 'Httpd.lns')
    aug.set('/augeas/load/Httpd/incl[last() + 1]', conf_file)
    aug.load()

    aug.defvar('conf', '/files' + conf_file)

    for match in aug.match('/files' + conf_file +
                           '/directive["RedirectMatch"]'):
        if aug.get(match + "/arg[1]") == '''"^/$"''':
            return aug.get(match + "/arg[2]").strip('"')

    return None


def get_home_page():
    """Return the shortcut ID that is set as current home page."""
    url = _get_home_page_url()
    return home_page_url2scid(url)


def change_home_page(shortcut_id: str):
    """Change the FreedomBox's default redirect to URL of a shortcut."""
    url = _home_page_scid2url(shortcut_id)
    if url:
        url = str(url)

    # URL may be a reverse_lazy() proxy
    privileged.set_home_page(url)


def get_advanced_mode():
    """Get whether option is enabled."""
    from plinth import kvstore
    return kvstore.get_default(ADVANCED_MODE_KEY, False)


def set_advanced_mode(advanced_mode):
    """Turn on/off advanced mode."""
    from plinth import kvstore
    kvstore.set(ADVANCED_MODE_KEY, advanced_mode)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Privacy App."""

import pathlib

import augeas

from plinth.actions import privileged

CONFIG_FILE = pathlib.Path('/etc/popularity-contest.d/freedombox.conf')


@privileged
def setup():
    """Create initial popcon configuration."""
    CONFIG_FILE.parent.mkdir(exist_ok=True)
    CONFIG_FILE.touch()

    aug = _load_augeas()
    aug.set('ENCRYPT', 'yes')
    aug.save()

    # Set the vendor to 'FreedomBox' with 'Debian' as parent
    default_link = pathlib.Path('/etc/dpkg/origins/default')
    debian_link = pathlib.Path('/etc/dpkg/origins/debian')
    if default_link.is_symlink() and default_link.resolve() == debian_link:
        default_link.unlink()
        default_link.symlink_to('freedombox')


@privileged
def set_configuration(enable_popcon: bool | None = None):
    """Update popcon configuration."""
    aug = _load_augeas()
    if enable_popcon:
        aug.set('PARTICIPATE', 'yes')
    else:
        aug.set('PARTICIPATE', 'no')

    aug.save()


def get_configuration() -> dict[str, bool]:
    """Return if popcon participation is enabled."""
    aug = _load_augeas()
    value = aug.get('PARTICIPATE')
    return {'enable_popcon': (value == 'yes')}


def _load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Shellvars', str(CONFIG_FILE))
    aug.set('/augeas/context', '/files' + str(CONFIG_FILE))
    aug.load()
    return aug



# SPDX-License-Identifier: AGPL-3.0-or-later
"""App component for other apps to authenticate with OpenID Connect."""

from plinth import app as app_module


class OpenIDConnect(app_module.FollowerComponent):
    """Component to authentication with OpenID Connection."""

    def __init__(self, component_id: str, client_id: str, name: str,
                 redirect_uris: list[str], skip_authorization: bool = False):
        """Initialize the OpenID Connect component.

        name is a description name for the client that is only used for record
        keeping.

        client_id is ID to use when registring the app as client (relying
        party) with FreedomBox's OpenID Connect Provider.

        redirect_uris is list of string containing URIs that the user's agent
        may be redirected to after successful authentication. If the URIs
        contain {domain} in their string, it will be expanded to the list of
        all domains configured in FreedomBox.
        """
        super().__init__(component_id)

        self.client_id = client_id
        self.name = name
        self.redirect_uris = redirect_uris
        self.post_logout_redirect_uris = ''  # None is not allowed
        self._client_type = None
        self._authorization_grant_type = None
        self._algorithm = None
        self.hash_client_secret = False
        self.skip_authorization = skip_authorization

    @property
    def client_type(self):
        """Return the client type.

        This is a property instead of a simple attribute to avoid importing
        Application model during component.__init__(). This would require
        Django to be configured.
        """
        from oauth2_provider.models import Application
        return self._client_type or Application.CLIENT_CONFIDENTIAL

    @client_type.setter
    def client_type(self, value):
        """Set the client type."""
        self._client_type = value

    @property
    def authorization_grant_type(self):
        """Return the authorization grant type.

        This is a property instead of a simple attribute to avoid importing
        Application model during component.__init__(). This would require
        Django to be configured.
        """
        from oauth2_provider.models import Application
        return (self._authorization_grant_type
                or Application.GRANT_AUTHORIZATION_CODE)

    @authorization_grant_type.setter
    def authorization_grant_type(self, value):
        """Set the authorization grant type."""
        self._authorization_grant_type = value

    @property
    def algorithm(self):
        """Return the algorithm.

        This is a property instead of a simple attribute to avoid importing
        Application model during component.__init__(). This would require
        Django to be configured.
        """
        from oauth2_provider.models import Application
        return self._algorithm or Application.HS256_ALGORITHM

    @algorithm.setter
    def algorithm(self, value):
        """Set the algorithm."""
        self._algorithm = value

    def get_client_secret(self):
        """Return the client secret stored for the application."""
        from oauth2_provider.models import Application
        return Application.objects.get_by_natural_key(
            self.client_id).client_secret

    def setup(self, old_version: int) -> None:
        """Register the app as client."""
        self._create_or_update_application()

    @staticmethod
    def update_domains_for_all():
        """For all app components, update redirect URIs and allowed origins."""
        for app in app_module.App.list():
            for component in app.components.values():
                if isinstance(component, OpenIDConnect):
                    component._create_or_update_application()

    def _create_or_update_application(self) -> None:
        """Register the app as client."""
        from oauth2_provider.models import Application
        try:
            application = Application.objects.get_by_natural_key(
                self.client_id)
            self._update_application(application)
        except Application.DoesNotExist:
            self._create_application()

    def _create_application(self) -> None:
        """Create a new application object."""
        from oauth2_provider import generators
        from oauth2_provider.models import Application
        client_secret = generators.generate_client_secret()
        Application.objects.create(
            client_id=self.client_id, client_secret=client_secret, user=None,
            redirect_uris=self._get_redirect_uris(),
            post_logout_redirect_uris=self.post_logout_redirect_uris,
            client_type=self.client_type,
            authorization_grant_type=self.authorization_grant_type,
            hash_client_secret=self.hash_client_secret, name=str(self.name),
            algorithm=self.algorithm,
            allowed_origins=self._get_allowed_origins(),
            skip_authorization=self.skip_authorization)

    def _update_application(self, application) -> None:
        """Update configuration for an existing application."""
        application.user = None
        application.redirect_uris = self._get_redirect_uris()
        application.post_logout_redirect_uris = self.post_logout_redirect_uris
        application.client_type = self.client_type
        application.authorization_grant_type = self.authorization_grant_type
        application.hash_client_secret = self.hash_client_secret
        application.name = str(self.name)
        application.algorithm = self.algorithm
        application.allowed_origins = self._get_allowed_origins()
        application.skip_authorization = self.skip_authorization
        application.save()

    def _get_redirect_uris(self) -> str:
        """Return an expanded list of redirect URIs."""
        from plinth.modules.names.components import DomainName
        final_uris = []
        # redirect_uris list can't be empty. Otherwise, validations for
        # 'localhost' and IP addresses won't work.
        domains = set(DomainName.list_names()) | {'localhost'}
        for uri in self.redirect_uris:
            if '{domain}' in uri:
                for domain in domains:
                    final_uris.append(uri.format(domain=domain))
            else:
                final_uris.append(uri)

        return ' '.join(final_uris)

    def _get_allowed_origins(self) -> str:
        """Return a list of all allowed origins for CORS header."""
        from plinth.modules.names.components import DomainName

        # redirect_uris list can't be empty. Otherwise, validations for
        # 'localhost' and IP addresses won't work. Keep origins in line with
        # redirect_uris.
        domains = set(DomainName.list_names()) | {'localhost'}
        origins = [f'https://{domain_name}' for domain_name in domains]
        return ' '.join(origins)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Django middleware to redirect to firstboot wizard if it has not be run
yet.
"""

import logging

from django.conf import settings
from django.http.response import HttpResponseRedirect
from django.urls import reverse
from django.utils.deprecation import MiddlewareMixin

from plinth import setup
from plinth.modules import first_boot
from plinth.utils import is_user_admin

LOGGER = logging.getLogger(__name__)


class FirstBootMiddleware(MiddlewareMixin):
    """Forward to firstboot page if firstboot isn't finished yet."""

    @staticmethod
    def process_request(request):
        """Handle a request as Django middleware request handler."""
        # Don't interfere with login page
        user_requests_login = request.path.startswith(
            reverse(settings.LOGIN_URL))
        if user_requests_login:
            return

        # Don't interfere with help or status pages
        user_requests_help = request.path.startswith(reverse('help:index'))
        user_requests_status = request.path.startswith(reverse('status'))
        if user_requests_help or user_requests_status:
            return

        # Don't interfere with first setup progress page. When first setup is
        # still running, no apps may have provided the first boot steps. This
        # will result in first boot wizard getting marked as completed
        # prematurely.
        if setup.is_first_setup_running:
            return

        firstboot_completed = first_boot.is_completed()
        user_requests_firstboot = first_boot.is_firstboot_url(request.path)

        # If user requests a step other than the welcome step, verify that they
        # indeed completed the secret verification by looking at the session.
        if (user_requests_firstboot
                and not request.path.startswith(reverse('first_boot:welcome'))
                and first_boot.firstboot_wizard_secret_exists()
                and not request.session.get('firstboot_secret_provided', False)
                and not is_user_admin(request)):
            return HttpResponseRedirect(reverse('first_boot:welcome'))

        # Redirect to first boot if requesting normal page and first
        # boot is not complete.
        if not firstboot_completed and not user_requests_firstboot:
            next_step = first_boot.next_step_or_none()
            if next_step:
                return HttpResponseRedirect(reverse(next_step))
            else:
                # No more steps in first boot
                first_boot.set_completed()

        # Redirect to 'complete' page if user requested firstboot after it is
        # finished.
        if firstboot_completed and user_requests_firstboot:
            return HttpResponseRedirect(reverse('first_boot:complete'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Forms for radicale module.
"""

from django import forms
from django.utils.translation import gettext_lazy as _

from plinth import cfg
from plinth.utils import format_lazy

CHOICES = [
    ('owner_only',
     _('Only the owner of a calendar/addressbook can view or '
       'make changes.')),
    ('owner_write',
     format_lazy(
         _('Any user with a {box_name} login can view any calendar/addressbook'
           ', but only the owner can make changes.'),
         box_name=_(cfg.box_name))),
    ('authenticated',
     format_lazy(
         _('Any user with a {box_name} login can view or make changes'
           ' to any calendar/addressbook.'), box_name=_(cfg.box_name))),
]


class RadicaleForm(forms.Form):
    """Specialized configuration form for radicale service."""
    access_rights = forms.ChoiceField(label=_('Access rights'),
                                      choices=CHOICES, required=True,
                                      widget=forms.RadioSelect())



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Utilities to help with PostgreSQL databases.

Uses utilities from 'postgres' package such as 'psql' and 'pg_dump'.
"""

import os
import pathlib

from plinth import action_utils


def _run_as(command, **kwargs):
    """Run a command as 'postgres' user."""
    command = ['sudo', '--user', 'postgres'] + command
    return action_utils.run(command, check=True, **kwargs)


def run_query(query):
    """Run a database query as 'postgres' user.

    Does not ensure that database server is running.
    """
    env = os.environ.copy()
    env['ON_ERROR_EXIT'] = '1'
    return _run_as(['psql', '--echo-errors'], env=env,
                   input=query.encode('utf-8'))


def _create_user(database_user: str, database_password: str):
    """Create a new user account with given credentials.

    Ignore errors if user already exists. Set password on the account either
    way. Passwords must be alphanumeric.
    """
    query = f'''
DO $$
BEGIN
    CREATE ROLE {database_user} WITH
        PASSWORD '{database_password}'
        NOSUPERUSER NOCREATEDB NOCREATEROLE INHERIT LOGIN NOREPLICATION
        NOBYPASSRLS;
EXCEPTION WHEN duplicate_object THEN
    ALTER ROLE {database_user} WITH
        PASSWORD '{database_password}';
END
$$;'''
    run_query(query)


def _drop_user(database_user: str):
    """Remove a user account with given username."""
    run_query(f'DROP ROLE {database_user};')


def create_database(database_name: str, database_user: str,
                    database_password: str):
    """Create a new database and a user account to access it.

    Database server is temporarily started if it is not running.
    """
    query = f'''
CREATE EXTENSION IF NOT EXISTS dblink;
DO $$
BEGIN
    PERFORM dblink_exec('',
        'CREATE DATABASE {database_name} WITH OWNER {database_user}');
EXCEPTION WHEN duplicate_database THEN
    ALTER DATABASE {database_name}
        OWNER TO {database_user};
END
$$;'''
    with action_utils.service_ensure_running('postgresql'):
        _create_user(database_user, database_password)
        run_query(query)


def drop_database(database_name: str, database_user: str):
    """Delete the database and the user account owning it.

    Database server is temporarily started if it is not running.
    """
    query = f'DROP DATABASE {database_name};'
    with action_utils.service_ensure_running('postgresql'):
        run_query(query)
        _drop_user(database_user)


def dump_database(backup_file: str | pathlib.Path, database_name: str):
    """Dump PostgreSQL database to a file.

    Database server is temporarily started if it is not running. Overwrite
    file if it exists.
    """
    backup_path = pathlib.Path(backup_file)
    with action_utils.service_ensure_running('postgresql'):
        with open(backup_path, 'w', encoding='utf-8') as file_handle:
            _run_as([
                'pg_dump', '--create', '--clean', '--if-exists', database_name
            ], stdout=file_handle)


def restore_database(backup_file: str | pathlib.Path, database_name: str,
                     database_user: str, database_password: str):
    """Restore database from a file.

    Database server is temporarily started if it is not running. User account
    is removed and recreated if it already exists. Drop database and recreate
    if it already exists.
    """
    with action_utils.service_ensure_running('postgresql'):
        drop_database(database_name, database_user)
        create_database(database_name, database_user, database_password)
        with open(backup_file, 'r', encoding='utf-8') as file_handle:
            _run_as(['psql', '--dbname', database_name], stdin=file_handle)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Let's Encrypt."""

import filecmp
import importlib
import inspect
import os
import pathlib
import re
import shutil
import sys
from typing import Any

from plinth import action_utils
from plinth import app as app_module
from plinth import module_loader
from plinth.actions import privileged
from plinth.modules import letsencrypt as le

TEST_MODE = False
LE_DIRECTORY = '/etc/letsencrypt/'
ETC_SSL_DIRECTORY = '/etc/ssl/'
AUTHENTICATOR = 'webroot'
WEB_ROOT_PATH = '/var/www/html'


def _get_certificate_expiry(domain: str) -> str:
    """Return the expiry date of a certificate."""
    certificate_file = os.path.join(le.LIVE_DIRECTORY, domain, 'cert.pem')
    output = action_utils.run(
        ['openssl', 'x509', '-enddate', '-noout', '-in', certificate_file],
        check=True).stdout
    return output.decode().strip().split('=')[1]


def _get_modified_time(domain: str) -> int:
    """Return the last modified time of a certificate."""
    certificate_file = pathlib.Path(le.LIVE_DIRECTORY) / domain / 'cert.pem'
    return int(certificate_file.stat().st_mtime)


def _get_validity_status(domain: str) -> str:
    """Return validity status of a certificate; valid, revoked, expired."""
    output = action_utils.run(['certbot', 'certificates', '-d', domain],
                              check=True).stdout
    line = output.decode(sys.stdout.encoding)

    match = re.search(r'INVALID: (.*)\)', line)
    if match is not None:
        validity = match.group(1).lower()
    elif re.search('VALID', line) is not None:
        validity = 'valid'
    else:
        validity = 'unknown'

    return validity


def _get_status() -> dict[str, Any]:
    """Return Python dictionary of currently configured domains.

    Should be run as root, otherwise might yield a wrong, empty answer.
    """
    try:
        domains = os.listdir(le.LIVE_DIRECTORY)
    except OSError:
        domains = []

    domains = [
        domain for domain in domains
        if os.path.isdir(os.path.join(le.LIVE_DIRECTORY, domain))
    ]

    domain_status = {}
    for domain in domains:
        domain_status[domain] = {
            'certificate_available':
                True,
            'expiry_date':
                _get_certificate_expiry(domain),
            'web_enabled':
                action_utils.webserver_is_enabled(domain, kind='site'),
            'validity':
                _get_validity_status(domain),
            'lineage':
                str(pathlib.Path(le.LIVE_DIRECTORY) / domain),
            'modified_time':
                _get_modified_time(domain)
        }
    return domain_status


@privileged
def get_status() -> dict[str, dict]:
    """Return a dictionary of currently configured domains."""
    domain_status = _get_status()
    return {'domains': domain_status}


@privileged
def get_modified_time(domain: str) -> int:
    """Return the modified time of a certificate as integer."""
    return _get_modified_time(domain)


@privileged
def revoke(domain: str):
    """Disable a domain and revoke the certificate."""
    cert_path = pathlib.Path(le.LIVE_DIRECTORY) / domain / 'cert.pem'
    if cert_path.exists():
        command = [
            'certbot', 'revoke', '--non-interactive', '--domain', domain,
            '--cert-path',
            str(cert_path)
        ]
        if TEST_MODE:
            command.append('--staging')

        action_utils.run(command, check=True)

    action_utils.webserver_disable(domain, kind='site')


@privileged
def obtain(domain: str):
    """Obtain a certificate for a domain and setup website."""
    command = [
        'certbot', 'certonly', '--non-interactive', '--text', '--agree-tos',
        '--register-unsafely-without-email', '--domain', domain,
        '--authenticator', AUTHENTICATOR, '--webroot-path', WEB_ROOT_PATH,
        '--renew-by-default'
    ]
    if TEST_MODE:
        command.append('--staging')

    action_utils.run(command, check=True)


@privileged
def copy_certificate(managing_app: str, source_private_key: str,
                     source_certificate: str, private_key: str,
                     certificate: str, user_owner: str, group_owner: str):
    """Copy certificate from LE directory to daemon's directory.

    Set ownership and permissions as requested needed by the daemon.

    """
    source_private_key_path = pathlib.Path(source_private_key).resolve()
    _assert_source_directory(source_private_key_path)
    source_certificate_path = pathlib.Path(source_certificate).resolve()
    _assert_source_directory(source_certificate_path)

    private_key_path = pathlib.Path(private_key).resolve()
    _assert_managed_path(managing_app, private_key_path)
    certificate_path = pathlib.Path(certificate).resolve()
    _assert_managed_path(managing_app, certificate_path)

    # Create directories, owned by root
    private_key_path.parent.mkdir(mode=0o755, parents=True, exist_ok=True)
    certificate_path.parent.mkdir(mode=0o755, parents=True, exist_ok=True)

    # Private key is only accessible to the user owner
    with action_utils.umask(0o177):
        shutil.copyfile(source_private_key_path, private_key_path)

        if certificate_path != private_key_path:
            # Certificate is only writable by the user owner
            with action_utils.umask(0o133):
                shutil.copyfile(source_certificate_path, certificate_path)
        else:
            # If private key and certificate are the same file, append one
            # after the other.
            source_certificate_bytes = source_certificate_path.read_bytes()
            with private_key_path.open(mode='a+b') as file_handle:
                file_handle.write(source_certificate_bytes)

    shutil.chown(certificate_path, user=user_owner, group=group_owner)
    shutil.chown(private_key_path, user=user_owner, group=group_owner)


@privileged
def compare_certificate(managing_app: str, source_private_key: str,
                        source_certificate: str, private_key: str,
                        certificate: str) -> bool:
    """Compare LE certificate with an app certificate."""
    source_private_key_path = pathlib.Path(source_private_key)
    source_certificate_path = pathlib.Path(source_certificate)
    _assert_source_directory(source_private_key_path)
    _assert_source_directory(source_certificate_path)

    private_key_path = pathlib.Path(private_key)
    certificate_path = pathlib.Path(certificate)
    _assert_managed_path(managing_app, private_key_path)
    _assert_managed_path(managing_app, certificate_path)

    result = False
    try:
        if filecmp.cmp(source_certificate_path, certificate_path) and \
           filecmp.cmp(source_private_key_path, private_key_path):
            result = True
    except FileNotFoundError:
        result = False

    return result


def _assert_source_directory(path):
    """Assert that a path is a valid source of a certificates."""
    assert (str(path).startswith(LE_DIRECTORY)
            or str(path).startswith(ETC_SSL_DIRECTORY))


def _get_managed_path(path):
    """Return the managed path given a certificate path."""
    if '{domain}' in path:
        return pathlib.Path(path.partition('{domain}')[0])

    return pathlib.Path(path).parent


def _assert_managed_path(module, path):
    """Check that path is in fact managed by module."""
    module_path = module_loader.get_module_import_path(module)
    module = importlib.import_module(module_path)
    module_classes = inspect.getmembers(module, inspect.isclass)
    app_classes = [
        cls[1] for cls in module_classes if issubclass(cls[1], app_module.App)
    ]

    managed_paths = []
    for cls in app_classes:
        app = cls()
        from plinth.modules.letsencrypt.components import LetsEncrypt
        components = app.get_components_of_type(LetsEncrypt)
        for component in components:
            if component.private_key_path:
                managed_paths.append(
                    _get_managed_path(component.private_key_path))
            if component.certificate_path:
                managed_paths.append(
                    _get_managed_path(component.certificate_path))

    if not set(path.parents).intersection(set(managed_paths)):
        raise AssertionError('Not a managed path')


@privileged
def delete(domain: str):
    """Disable a domain and delete the certificate."""
    command = ['certbot', 'delete', '--non-interactive', '--cert-name', domain]
    action_utils.run(command, check=True)
    action_utils.webserver_disable(domain, kind='site')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django forms for configuring TiddlyWiki."""

from django import forms
from django.core import validators
from django.utils.translation import gettext_lazy as _


def _validate_not_index_file(name):
    """Validate that the normalized file name is not 'index.html'."""
    if str(name) in ('index.html', 'index'):
        raise forms.ValidationError(
            _('Wiki files cannot be named "index.html".'))


class CreateWikiForm(forms.Form):
    """Form to create a new wiki file."""

    name = forms.CharField(
        label=_('Name of the wiki file, with file extension ".html"'),
        strip=True, validators=[_validate_not_index_file], help_text=_(
            'Wiki title and description can be set from within the wiki. '
            'This file name is independent of the wiki title.'))


class RenameWikiForm(forms.Form):
    """Form to rename a wiki file."""

    new_name = forms.CharField(
        label=_('New name for the wiki file, with file extension ".html"'),
        strip=True, validators=[_validate_not_index_file], help_text=_(
            'Renaming the file has no effect on the title of the wiki.'))


class UploadWikiForm(forms.Form):
    """Form to upload a wiki file."""

    file = forms.FileField(
        label=_('A TiddlyWiki file with .html file extension'), required=True,
        validators=[
            validators.FileExtensionValidator(
                ['html'],
                _('TiddlyWiki files must be in HTML format'),
            ), _validate_not_index_file
        ],
        help_text=_('Upload an existing TiddlyWiki file from this computer.'))



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Django context processors to provide common data to templates.
"""

from django.utils.translation import gettext as _
from django.utils.translation import gettext_noop

from plinth import cfg, views, web_server
from plinth.utils import is_user_admin


def common(request):
    """Add additional context values to RequestContext for use in templates.

    Any resources referenced in the return value are expected to have been
    initialized or configured externally beforehand.
    """
    # Allow a value in configuration file to be translated.  Allow
    # the brand name 'FreedomBox' itself to be translated.
    gettext_noop('FreedomBox')

    from plinth.notification import Notification
    notifications_context = Notification.get_display_context(
        request, user=request.user)

    breadcrumbs = views.get_breadcrumbs(request)
    active_section_url = [
        key for key, value in breadcrumbs.items()
        if value.get('is_active_section')
    ][0]
    return {
        'cfg': cfg,
        'breadcrumbs': breadcrumbs,
        'active_section_url': active_section_url,
        'box_name': _(cfg.box_name),
        'user_is_admin': is_user_admin(request, True),
        'user_css': web_server.get_user_css(),
        'notifications': notifications_context['notifications'],
        'notifications_max_severity': notifications_context['max_severity']
    }



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Application manifest for GNOME."""

from django.utils.translation import gettext_lazy as _

backup: dict = {}

tags = [
    _('Desktop'),
    _('Browser'),
    _('Office suite'),
    _('Software store'),
    _('GUI'),
    _('Graphical apps')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app to configure Nextcloud."""

import contextlib

from django.utils.translation import gettext_lazy as _

from plinth import app as app_module
from plinth import cfg, frontpage, menu
from plinth.config import DropinConfigs
from plinth.daemon import Daemon, SharedDaemon
from plinth.modules.apache.components import (Webserver, diagnose_url,
                                              diagnose_url_on_all)
from plinth.modules.backups.components import BackupRestore
from plinth.modules.firewall.components import (Firewall,
                                                FirewallLocalProtection)
from plinth.modules.names.components import DomainName
from plinth.package import Packages
from plinth.signals import domain_added, domain_removed
from plinth.utils import format_lazy

from . import manifest, privileged

_description = [
    _('Nextcloud is a self-hosted productivity platform which provides '
      'private and secure functions for file sharing, collaborative work, '
      'and more. Nextcloud includes the Nextcloud server, client applications '
      'for desktop computers, and mobile clients. The Nextcloud server '
      'provides a well integrated web interface.'),
    _('All users of FreedomBox can use Nextcloud. To perform administrative '
      f'actions, use the <strong>"{privileged.GUI_ADMIN}"</strong> user after '
      'setting a password here.'),
    format_lazy(
        _('Please note that Nextcloud is installed and run inside a container '
          'provided by the Nextcloud community. Security, quality, privacy '
          'and legal reviews are done by the upstream project and not by '
          'Debian/{box_name}. Updates are performed following an independent '
          'cycle.'), box_name=_(cfg.box_name)),
]


class NextcloudApp(app_module.App):
    """FreedomBox app for Nextcloud."""

    app_id = 'nextcloud'

    _version = 2

    configure_when_disabled = False

    def __init__(self):
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               name=_('Nextcloud'), icon_filename='nextcloud',
                               description=_description,
                               manual_page='Nextcloud',
                               clients=manifest.clients, tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-nextcloud', info.name, info.icon_filename,
                              info.tags, 'nextcloud:index',
                              parent_url_name='apps')
        self.add(menu_item)

        shortcut = frontpage.Shortcut('shortcut-nextcloud', info.name,
                                      icon=info.icon_filename,
                                      url='/nextcloud/', clients=info.clients,
                                      tags=info.tags, login_required=True)
        self.add(shortcut)

        packages = Packages('packages-nextcloud', [
            'podman', 'default-mysql-server', 'python3-iso3166',
            'redis-server', 'php-redis'
        ], conflicts=['libpam-tmpdir'],
                            conflicts_action=Packages.ConflictsAction.REMOVE)
        self.add(packages)

        dropin_configs = DropinConfigs('dropin-configs-nextcloud', [
            '/etc/apache2/conf-available/nextcloud-freedombox.conf',
            '/etc/redis/conf.d/freedombox.conf',
        ])
        self.add(dropin_configs)

        firewall = Firewall('firewall-nextcloud', info.name,
                            ports=['http', 'https'], is_external=True)
        self.add(firewall)

        firewall_local_protection = FirewallLocalProtection(
            'firewall-local-protection-nextcloud', ['9000'])
        self.add(firewall_local_protection)

        webserver = Webserver('webserver-nextcloud', 'nextcloud-freedombox')
        self.add(webserver)

        daemon = SharedDaemon('shared-daemon-podman-auto-update',
                              'podman-auto-update.timer')
        self.add(daemon)

        daemon = SharedDaemon('shared-daemon-nextcloud-redis', 'redis-server',
                              listen_ports=[(6379, 'tcp4')])
        self.add(daemon)

        daemon = SharedDaemon('shared-daemon-nextcloud-mysql', 'mysql')
        self.add(daemon)

        daemon = NextcloudDaemon('daemon-nextcloud', 'nextcloud-freedombox')
        self.add(daemon)

        daemon = Daemon('daemon-nextcloud-timer',
                        'nextcloud-cron-freedombox.timer')
        self.add(daemon)

        backup_restore = NextcloudBackupRestore('backup-restore-nextcloud',
                                                **manifest.backup)
        self.add(backup_restore)

    @staticmethod
    def post_init():
        """Perform post initialization operations."""
        domain_added.connect(_on_domain_added)
        domain_removed.connect(_on_domain_removed)

    def setup(self, old_version):
        """Install and configure the app."""
        if old_version == 1:
            with _ensure_nextcloud_running():
                # Reset the value so that 'overwrite.cli.url' is set to default
                # value instead of it being missing.
                if not privileged.get_override_domain():
                    privileged.set_override_domain('')

        super().setup(old_version)
        # Drop-in configs need to be enabled for setup to succeed
        self.get_component('dropin-configs-nextcloud').enable()
        redis = self.get_component('shared-daemon-nextcloud-redis')
        mysql = self.get_component('shared-daemon-nextcloud-mysql')
        nextcloud = self.get_component('daemon-nextcloud')

        # Determine whether app should be disabled after setup
        should_disable = old_version and not nextcloud.is_enabled()

        with redis.ensure_running():
            with mysql.ensure_running():
                # Database needs to be running for successful initialization or
                # upgrade of Nextcloud database.
                privileged.setup()
                _set_trusted_domains()

        if should_disable:
            self.disable()

        if not old_version:
            self.enable()

    def uninstall(self):
        """De-configure and uninstall the app."""
        privileged.uninstall()
        super().uninstall()

    def diagnose(self):
        """Run diagnostics and return the results.

        When an override domain is set, that domain and all other addresses are
        expected to work. This is because Nextcloud will accept any Host: HTTP
        header and then override it with the provided domain name. When
        override domain is not set, only the configured trusted domains along
        with local IP addresses are allowed. Others are rejected with an error.
        """
        results = super().diagnose()

        kwargs = {'check_certificate': False}
        url = 'https://{domain}/nextcloud/login'
        domain = privileged.get_override_domain()
        if domain:
            results.append(diagnose_url(url.format(domain=domain), **kwargs))
            results += diagnose_url_on_all(url.format(domain='{host}'),
                                           **kwargs)
        else:
            local_addresses = [('localhost', '4'), ('localhost', '6'),
                               ('127.0.0.1', '4'), ('[::1]', '6')]
            for address, kind in local_addresses:
                results.append(
                    diagnose_url(url.format(domain=address), kind=kind,
                                 **kwargs))

        return results


class NextcloudDaemon(Daemon):
    """Component to manage Nextcloud container service."""

    def is_enabled(self):
        """Return if the daemon/unit is enabled."""
        return privileged.is_enabled()

    def enable(self):
        """Run operations to enable the daemon/unit."""
        super().enable()
        privileged.enable()

    def disable(self):
        """Run operations to disable the daemon/unit."""
        super().disable()
        privileged.disable()


class NextcloudBackupRestore(BackupRestore):
    """Component to backup/restore Nextcloud."""

    def backup_pre(self, packet):
        """Save database contents."""
        super().backup_pre(packet)
        with _ensure_nextcloud_running():
            privileged.dump_database()

    def restore_post(self, packet):
        """Restore database contents."""
        super().restore_post(packet)
        with _ensure_nextcloud_running():
            privileged.restore_database()


def _on_domain_added(sender, domain_type, name='', description='',
                     services=None, **kwargs):
    """Add domain to list of trusted domains."""
    app = app_module.App.get('nextcloud')
    if app.needs_setup():
        return

    _set_trusted_domains()


def _on_domain_removed(sender, domain_type, name='', **kwargs):
    """Update the list of trusted domains."""
    app = app_module.App.get('nextcloud')
    if app.needs_setup():
        return

    _set_trusted_domains()


def _set_trusted_domains() -> None:
    """Set the list of trusted domains."""
    all_domains = DomainName.list_names()
    with _ensure_nextcloud_running():
        privileged.set_trusted_domains(all_domains)


@contextlib.contextmanager
def _ensure_nextcloud_running():
    """Ensure the nextcloud is running and returns to original state."""
    app = app_module.App.get('nextcloud')
    app.get_component('dropin-configs-nextcloud').enable()
    mysql = app.get_component('shared-daemon-nextcloud-mysql')
    redis = app.get_component('shared-daemon-nextcloud-redis')
    container = app.get_component('daemon-nextcloud')
    with mysql.ensure_running():
        with redis.ensure_running():
            with container.ensure_running():
                yield



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Radicale."""

import os
import shutil

import augeas

from plinth import action_utils
from plinth.actions import privileged

CONFIG_FILE = '/etc/radicale/config'
LOG_PATH = '/var/log/radicale'
SERVICE_NAME = 'uwsgi-app@radicale.service'


@privileged
def setup():
    """Configure authentication to Apache remote user.

    For radicale < 3.5 (bookworm), this was set by configuration shipped with
    Debian. For radicale >= 3.5 (trixie), this needs to be explicitly. For
    simplicity, set is unconditionally.
    """
    aug = load_augeas()
    aug.set('auth/type', 'remote_user')
    aug.set('auth/lc_username', 'True')
    aug.save()
    # Service is started again by socket.
    action_utils.service_stop(SERVICE_NAME)


@privileged
def configure(rights_type: str):
    """Set the radicale rights type to a particular value."""
    if rights_type == 'owner_only':
        # Default rights file is equivalent to owner_only.
        rights_type = 'from_file'

    aug = load_augeas()
    aug.set('rights/type', rights_type)
    aug.save()
    action_utils.service_stop(SERVICE_NAME)


@privileged
def fix_paths():
    """Fix log path to work around a bug."""
    # Workaround for bug in radicale's uwsgi script (#931201)
    if not os.path.exists(LOG_PATH):
        os.makedirs(LOG_PATH)


def load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    # INI file lens
    aug.transform('Radicale', CONFIG_FILE)
    aug.set('/augeas/context', '/files' + CONFIG_FILE)
    aug.load()
    return aug


def get_rights_value():
    """Returns the current Rights value."""
    aug = load_augeas()
    value = aug.get('rights/type')

    if value == 'from_file':
        # Default rights file is equivalent to owner_only.
        value = 'owner_only'

    return value


@privileged
def uninstall():
    """Remove all radicale collections."""
    shutil.rmtree('/var/lib/private/radicale/', ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""URLs for the Apache module."""

from django.urls import re_path
from stronghold.decorators import public

from .views import DiscoverIDPView

urlpatterns = [
    re_path(r'^apache/discover-idp/$', public(DiscoverIDPView.as_view()),
            name='discover-idp'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for the Coturn app component.
"""

import json
from unittest.mock import call, patch

import pytest

from plinth.utils import random_string

from .. import notify_configuration_change
from ..components import (TurnConfiguration, TurnConsumer,
                          TurnTimeLimitedConsumer, UserTurnConfiguration)


@pytest.fixture(name='turn_configuration')
def fixture_turn_configuration():
    """Return test Coturn configuration."""
    return TurnConfiguration('test-domain.example', [], random_string(64))


@pytest.fixture(name='empty_component_list', autouse=True)
def fixture_empty_component_list():
    """Remove all entries in component list before starting a test."""
    TurnConsumer._all = {}


def test_turn_configuration_init():
    """Test creating configuration object."""
    config = TurnConfiguration('test-domain.example', [], 'test-shared-secret')
    assert config.domain == 'test-domain.example'
    assert config.shared_secret == 'test-shared-secret'
    assert config.uris == [
        'stun:test-domain.example:3478',
        'turn:test-domain.example:3478?transport=tcp',
        'turn:test-domain.example:3478?transport=udp',
    ]

    config = TurnConfiguration(None, ['test-uri1', 'test-uri2'],
                               'test-shared-secret')
    assert config.domain is None
    assert config.uris == ['test-uri1', 'test-uri2']

    config = TurnConfiguration('test-domain.example',
                               ['test-uri1', 'test-uri2'],
                               'test-shared-secret')
    assert config.domain == 'test-domain.example'
    assert config.uris == ['test-uri1', 'test-uri2']

    config = UserTurnConfiguration('test-domain.example',
                                   ['test-uri1', 'test-uri2'], None,
                                   'test-username', 'test-credential')
    assert config.domain == 'test-domain.example'
    assert config.uris == ['test-uri1', 'test-uri2']
    assert config.shared_secret is None
    assert config.username == 'test-username'
    assert config.credential == 'test-credential'


def test_turn_configuration_to_json():
    """Test exporting configuration to JSON."""
    config = TurnConfiguration('test-domain.example', [], 'test-shared-secret')
    assert json.loads(config.to_json()) == {
        'domain': 'test-domain.example',
        'uris': [
            'stun:test-domain.example:3478',
            'turn:test-domain.example:3478?transport=tcp',
            'turn:test-domain.example:3478?transport=udp',
        ],
        'shared_secret': 'test-shared-secret'
    }


def test_turn_configuration_validate_turn_uris():
    """Test validation method to check for STUN/TURN URIs."""
    valid_uris = [
        'stun:test-domain.example:3478',
        'turn:test-domain.example:3478?transport=tcp',
        'turn:test-domain.example:3478?transport=udp',
    ]
    invalid_uris = [
        'x-invalid-uri',
        'stun:',
        'stun:domain-port-missing.example',
        'stun:testing.example:1234invalid-append',
        'turn:testing.example:1234',
        'turn:testing.example:1234?invalid-param=value',
        'turn:testing.example:1234?transport=invalid-value',
        'turn:testing.example:1234?transport=tcp-invalid-append',
    ]
    assert TurnConfiguration().validate_turn_uris(valid_uris)
    for uri in invalid_uris:
        assert not TurnConfiguration().validate_turn_uris([uri])


def test_component_init_and_list():
    """Test initializing and listing all the components."""
    component1 = TurnConsumer('component1')
    component2 = TurnConsumer('component2')
    component3 = TurnTimeLimitedConsumer('component3')
    assert component1.component_id == 'component1'
    assert [component1, component2, component3] == list(TurnConsumer.list())


@patch('plinth.modules.coturn.get_config')
def test_notify_on_configuration_changed(get_config, turn_configuration):
    """Test configuration change notifications."""
    component = TurnConsumer('component')
    get_config.return_value = turn_configuration
    with patch.object(component, 'on_config_change') as mock_method:
        notify_configuration_change()
        mock_method.assert_has_calls([call(turn_configuration)])


@patch('plinth.modules.coturn.get_config')
def test_get_configuration(get_config, turn_configuration):
    """Test coturn configuration retrieval using component."""
    get_config.return_value = turn_configuration
    component = TurnConsumer('component')
    assert component.get_configuration() == turn_configuration


@patch('plinth.modules.coturn.get_config')
def test_get_user_configuration(get_config, turn_configuration):
    """Test coturn user configuration retrieval using component."""
    get_config.return_value = turn_configuration
    component = TurnTimeLimitedConsumer('component')
    user_config = component.get_configuration()
    assert user_config.domain == turn_configuration.domain
    assert user_config.uris == turn_configuration.uris
    assert user_config.shared_secret is None
    assert user_config.username is not None
    assert user_config.credential is not None



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for database utilities.
"""
import threading
import time

from .. import db
from ..db import dbconfig


def test_db_lock_no_wait():
    """Test that lock is immediately by first user."""
    lock = db.DBLock()

    start_time = time.time()
    with lock:
        pass

    end_time = time.time()
    assert end_time - start_time < 0.1


def test_db_lock_max_wait():
    """Test that lock waits only for timeout period."""
    event = threading.Event()
    lock = db.DBLock()
    lock.timeout = 0.25

    def thread_func():
        with lock:
            event.set()
            time.sleep(0.3)

    thread = threading.Thread(target=thread_func)
    thread.start()

    event.wait()
    start_time = time.time()
    with lock as return_value:
        pass

    end_time = time.time()
    assert end_time - start_time < 0.27
    assert not return_value


def test_db_lock_release():
    """Test that lock is available after release."""
    event = threading.Event()
    lock = db.DBLock()
    lock.timeout = 0.25

    def thread_func():
        with lock:
            event.set()
            time.sleep(0.2)

    thread = threading.Thread(target=thread_func)
    thread.start()

    event.wait()
    start_time = time.time()
    with lock as return_value:
        pass

    end_time = time.time()
    assert return_value
    assert end_time - start_time <= 0.23


def test_dbconfig_get_credentials(tmp_path):
    """Test that parsing a dbconfig-common file works."""
    file_path = tmp_path / 'test.conf'
    configuration = '''
dbc_dbserver='localhost'
dbc_dbname='miniflux'
dbc_dbuser='miniflux'
dbc_dbpass='gCcNyWjyPjDH'
'''
    file_path.write_text(configuration)
    credentials = dbconfig.get_credentials(file_path)
    assert credentials == {
        'host': 'localhost',
        'database': 'miniflux',
        'user': 'miniflux',
        'password': 'gCcNyWjyPjDH',
    }



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URLs for the Zoph module.
"""

from django.urls import re_path

from .views import SetupView, ZophAppView

urlpatterns = [
    re_path(r'^apps/zoph/setup/$', SetupView.as_view(), name='setup'),
    re_path(r'^apps/zoph/$', ZophAppView.as_view(app_id='zoph'), name='index')
]
